#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Common client utilities"""

import argparse
import os


def env(*vars, **kwargs):
    """Search for the first defined of possibly many env vars

    Returns the first environment variable defined in vars, or
    returns the default defined in kwargs.
    """
    for v in vars:
        value = os.environ.get(v, None)
        if value:
            return value
    return kwargs.get('default', '')


class KeyValueAction(argparse.Action):
    """A custom action to parse arguments as key=value pairs. """

    def __call__(self, parser, namespace, values, option_string=None):
        # Make sure we have an empty dict rather than None
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})

        # Add value if an assignment else remove it
        if '=' in values:
            getattr(namespace, self.dest, {}).update([values.split('=', 1)])
        else:
            getattr(namespace, self.dest, {}).pop(values, None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.nodetemplate import NodeTemplate
from toscaparser.tests.base import TestCase
import toscaparser.utils.yamlparser
from translator.hot.tosca.tosca_floating import ToscaFloatingIP


class ToscaFloatingIPTest(TestCase):

    def _tosca_floatingip_test(self, tpl_snippet, expectedprops, name=None):
        nodetemplates = (toscaparser.utils.yamlparser.
                         simple_parse(tpl_snippet)['node_templates'])
        if not name:
            name = list(nodetemplates.keys())[0]
        nodetemplate = NodeTemplate(name, nodetemplates, custom_def=[])
        nodetemplate.validate()
        tosca_floatingip = ToscaFloatingIP(nodetemplate)
        tosca_floatingip.handle_properties()
        self.assertEqual(expectedprops, tosca_floatingip.properties)

    def test_node_floatingip_with_properties(self):
        tpl_snippet = '''
        node_templates:
          floating_ip:
            type: tosca.nodes.network.FloatingIP
            properties:
              floating_network: public
              floating_ip_address: 192.168.56.8
              port_id: abcd
        '''
        expectedprops = {'floating_network': 'public',
                         'floating_ip_address': '192.168.56.8',
                         'port_id': 'abcd'}
        self._tosca_floatingip_test(
            tpl_snippet,
            expectedprops)

    def test_node_floatingip_with_properties_and_link_requirements(self):
        tpl_snippet = '''
        node_templates:
          floating_ip:
            type: tosca.nodes.network.FloatingIP
            properties:
              floating_network: public
              floating_ip_address: 192.168.56.8
            requirements:
              - link:
                  node: port1
          port1:
            type: tosca.nodes.network.Port
            properties:
              ip_address: 10.0.0.6
        '''
        expectedprops = {'floating_network': 'public',
                         'floating_ip_address': '192.168.56.8',
                         'port_id': '{ get_resource: port1 }'}
        self._tosca_floatingip_test(
            tpl_snippet,
            expectedprops,
            name='floating_ip')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from translator.hot.syntax.hot_resource import HotResource
log = logging.getLogger('heat-translator')


# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaNfvVducp'
TOSCA_LINKS_TO = 'tosca.relationships.nfv.VirtualLinksTo'
TOSCA_BINDS_TO = 'tosca.relationships.nfv.VirtualBindsTo'
depends_on_set = (
    TOSCA_LINKS_TO,
    TOSCA_BINDS_TO,
)


class ToscaNfvVducp(HotResource):
    """Translate TOSCA node type tosca.nodes.nfv.VduCp."""

    toscatype = 'tosca.nodes.nfv.VduCp'
    EXTERNAL_VL = '#ADD_YOUR_NETWORK_HERE'

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaNfvVducp, self).__init__(
            nodetemplate,
            type='OS::Neutron::Port',
            csar_dir=csar_dir)

        # Default order
        self.order = 0

        # Extract virtual_link and virtual_binding
        self.virtual_link = None
        self.virtual_binding = None
        requirements = self.nodetemplate.requirements
        for req in requirements:
            if 'virtual_link' in req:
                vl_val = req.get('virtual_link')
                if isinstance(vl_val, dict):
                    self.virtual_link = vl_val.get('node')
                else:
                    self.virtual_link = vl_val
            elif 'virtual_binding' in req:
                self.virtual_binding = req.get('virtual_binding')

    def _generate_networks_for_compute(self, port_resources):
        '''Generate compute networks property list from the port resources.'''
        networks = []
        for resource in port_resources:
            networks.append({'port': '{ get_resource: %s }' % (resource.name)})
        return networks

    def _insert_sorted_resource(self, resources, resource):
        '''Insert a resource in the list of resources and keep the order.'''
        lo = 0
        hi = len(resources)
        while lo < hi:
            mid = (lo + hi) // 2
            if resource.order < resources[mid].order:
                hi = mid
            else:
                lo = mid + 1
        resources.insert(lo, resource)

    def handle_properties(self):
        tosca_props = self.get_tosca_props()
        own_props = {}

        if self.virtual_link:
            own_props['network'] = '{ get_resource: %s }' % (
                self.virtual_link
            )
        else:
            # If no virtual_binding is specified in VNFD, users
            # have to input external VL given in instantiation VNF request.
            own_props['network'] = '%s' % (
                self.EXTERNAL_VL
            )

        for key, value in tosca_props.items():
            if key == 'order':
                self.order = value
            elif key == 'vnic_type':
                own_props['binding:vnic_type'] = value

        # Remove depends_on
        self.remove_depends_on(depends_on_set)

        binds_to = None
        for rel, node in self.nodetemplate.relationships.items():
            # Check for BindsTo relationship. If found add network to the
            # network property of the corresponding compute resource
            if not binds_to and rel.is_derived_from(TOSCA_BINDS_TO):
                binds_to = node
                for hot_resource in self.depends_on_nodes:
                    if binds_to.name == hot_resource.name:
                        port_rsrcs = hot_resource.assoc_port_resources
                        self._insert_sorted_resource(port_rsrcs, self)
                        networks = \
                            self._generate_networks_for_compute(port_rsrcs)
                        hot_resource.properties['networks'] = networks
                        break

        self.properties = own_props



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from translator.hot.syntax.hot_resource import HotResource
from translator.hot.tosca.etsi_nfv.scalingaspect.\
    tosca_policies_nfv_scalingaspect import ToscaNfvScalingAspect
log = logging.getLogger('heat-translator')

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaNfvScalingAspects'


class ToscaNfvScalingAspects(HotResource):
    """Translate TOSCA policy type tosca.policies.nfv.ScalingAspects."""

    toscatype = 'tosca.policies.nfv.ScalingAspects'

    def __init__(self, policy, csar_dir=None, hot_template_parameters=None,
                 nested_template=None,
                 extra_flg=None):
        hot_type = "OS::Heat::ScalingPolicy"

        if extra_flg is None:
            # Multi aspects support
            self.aspect_names = policy.properties['aspects'].keys()
            self.scaling_aspect_objs = []
            for aspect_name in self.aspect_names:
                self.scaling_aspect_objs.append(
                    ToscaNfvScalingAspect(policy,
                                          aspect_name,
                                          csar_dir,
                                          hot_template_parameters)
                )

        super(ToscaNfvScalingAspects, self).__init__(
            policy,
            type=hot_type,
            csar_dir=csar_dir)

        self.policy = policy
        self.hot_template_parameters = hot_template_parameters
        self.multi_nested_templates = {}
        self.nested_template = \
            {} if nested_template is None else nested_template

    def handle_properties(self, resources):
        for aspect_obj in self.scaling_aspect_objs:
            resources, nstd_tmpt = aspect_obj.handle_properties(resources)
            self.multi_nested_templates[aspect_obj.aspect_name] = nstd_tmpt

        resources = self._create_scale_out_in_resources(resources)
        return resources

    def extract_substack_templates(self, base_filename, hot_template_version):
        return self.nested_template

    def _create_scale_out_in_resources(self, resources):
        for asp_obj in self.scaling_aspect_objs:
            asp_name = asp_obj.aspect_name

            # Create scale_out and scale_in resources
            scl_out_rsrc = ToscaNfvScalingAspects(
                self.policy,
                csar_dir=self.csar_dir,
                hot_template_parameters=self.
                hot_template_parameters,
                nested_template=self.
                multi_nested_templates[asp_name],
                extra_flg=True
            )
            scl_in_rsrc = ToscaNfvScalingAspects(
                self.policy,
                csar_dir=self.csar_dir,
                hot_template_parameters=self.
                hot_template_parameters,
                nested_template=self.
                multi_nested_templates[asp_name],
                extra_flg=True
            )
            scl_out_rsrc.name = asp_name + '_scale_out'
            scl_in_rsrc.name = asp_name + '_scale_in'

            if asp_obj.scaling_adjustment is not None:
                scaling_adjustment = \
                    int(asp_obj.scaling_adjustment)
                scl_out_rsrc.properties['scaling_adjustment'] = \
                    scaling_adjustment
                scl_in_rsrc.properties['scaling_adjustment'] = \
                    -1 * scaling_adjustment
            else:
                log.warning('No ScalingAspectDelta for %s of %s, %s is found'
                            % (asp_obj.vdu_name, asp_name, asp_obj.delta_name))

            scl_out_rsrc.properties["auto_scaling_group_id"] = {
                'get_resource': asp_name
            }
            scl_out_rsrc.properties["adjustment_type"] = \
                "change_in_capacity"
            resources.append(scl_out_rsrc)

            scl_in_rsrc.properties["auto_scaling_group_id"] = {
                'get_resource': asp_name
            }
            scl_in_rsrc.properties["adjustment_type"] = \
                "change_in_capacity"
            resources.append(scl_in_rsrc)

        # This resource has already created scale_out and
        # scale_in resources, so remove it.
        resources.remove(self)

        return resources



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Translate action implementations"""

import os
import sys

from cliff import command
from oslo_log import log as logging

from toscaparser.tosca_template import ToscaTemplate
from toscaparser.utils.gettextutils import _
from translator.common import flavors
from translator.common import images
from translator.common.utils import UrlUtils
from translator.hot.tosca_translator import TOSCATranslator
from translator.osc import utils

LOG = logging.getLogger('heat-translator')


class TranslateTemplate(command.Command):

    """Translate a template"""

    auth_required = True

    def get_parser(self, prog_name):
        parser = super(TranslateTemplate, self).get_parser(prog_name)
        parser.add_argument(
            '--template-file',
            metavar='<template-file>',
            required=True,
            help='Path to the file that needs to be translated.')
        parser.add_argument(
            '--template-type',
            metavar='<template-type>',
            required=True,
            choices=['tosca'],
            help='Format of the template file.')
        parser.add_argument(
            '--output-file',
            metavar='<output-file>',
            help='Path to place the translated content.')
        parser.add_argument(
            '--parameter',
            metavar='<key=value>',
            action=utils.KeyValueAction,
            help='Set a property for this template '
                 '(repeat option to set multiple properties)',
        )
        parser.add_argument(
            '--validate-only',
            metavar='<true or false>',
            help='Set to true to only validate a template file.',
            default='false')
        return parser

    def take_action(self, parsed_args):
        LOG.debug(_('Translating the template with input parameters'
                    '(%s).'), parsed_args)
        output = None

        session = self.app.cloud.get_session()
        flavors.SESSION = session
        images.SESSION = session

        if parsed_args.parameter:
            parsed_params = parsed_args.parameter
        else:
            parsed_params = {}

        if parsed_args.template_type == "tosca":
            path = parsed_args.template_file
            a_file = os.path.isfile(path)
            a_url = UrlUtils.validate_url(path) if not a_file else False
            if a_file or a_url:
                validate = parsed_args.validate_only
                if validate and validate.lower() == "true":
                    ToscaTemplate(path, parsed_params, a_file)
                    msg = (_('The input "%(path)s" successfully passed '
                             'validation.') % {'path': path})
                    print(msg)
                else:
                    tosca = ToscaTemplate(path, parsed_params, a_file)
                    translator = TOSCATranslator(tosca, parsed_params)
                    output = translator.translate()
            else:
                msg = _('Could not find template file.\n')
                LOG.error(msg)
                sys.stdout.write(msg)
                raise SystemExit

        if output:
            if parsed_args.output_file:
                with open(parsed_args.output_file, 'w+') as f:
                    f.write(output)
            else:
                print(output)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from translator.conf.config import ConfigProvider as translatorConfig
from translator.tests.base import TestCase


def reload_config(func):
    '''Decorator to reload config.

    Set to default values defined in translator.conf file

    '''

    def reload(*args):
        func(*args)
        path = os.path.dirname(os.path.abspath(__file__)) + '/../conf/'
        conf_file = os.path.join(path, 'translator.conf')
        translatorConfig._load_config(conf_file)

    return reload


class ConfTest(TestCase):

    @reload_config
    @mock.patch('configparser.ConfigParser')
    def test_load_config(self, mock_config_parser):
        translatorConfig._translator_config.read = mock.MagicMock()
        translatorConfig._load_config('fake_file.conf')
        self.assertTrue(translatorConfig._translator_config.read.called)

    @mock.patch.object(translatorConfig._translator_config, 'get')
    def test_get_value(self, mock_translator_config):
        mock_translator_config.return_value = 'hot'
        value = translatorConfig.get_value('DEFAULT', 'language')
        self.assertTrue(translatorConfig._translator_config.get.called)
        self.assertEqual(value, 'hot')

    @mock.patch.object(translatorConfig._translator_config, 'items')
    def test_get_all_values(self, mock_translator_config):
        mock_translator_config.return_value = ['hot']
        values = translatorConfig.get_all_values()
        self.assertTrue(translatorConfig._translator_config.items.called)
        self.assertEqual(values[0], 'hot')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from toscaparser.elements.scalarunit import ScalarUnit_Size
from translator.hot.syntax.hot_resource import HotResource

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaObjectStorage'


class ToscaObjectStorage(HotResource):
    '''Translate TOSCA node type tosca.nodes.ObjectStorage.'''

    toscatype = 'tosca.nodes.ObjectStorage'

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaObjectStorage, self).__init__(nodetemplate,
                                                 type='OS::Swift::Container',
                                                 csar_dir=csar_dir)
        pass

    def handle_properties(self):
        tosca_props = self.get_tosca_props()
        objectstore_props = {}
        container_quota = {}
        skip_check = False

        for key, value in tosca_props.items():
            if key == "name":
                objectstore_props["name"] = value
            elif key == "size" or key == "maxsize":
                # currently heat is not supporting dynamically increase
                # the container quota-size.
                # if both defined in tosca template, consider store_maxsize.
                if skip_check:
                    continue
                quota_size = None
                if "maxsize" in tosca_props.keys():
                    quota_size = tosca_props["maxsize"]
                else:
                    quota_size = tosca_props["size"]
                container_quota["Quota-Bytes"] = \
                    ScalarUnit_Size(quota_size).get_num_from_scalar_unit()
                objectstore_props["X-Container-Meta"] = container_quota
                skip_check = True

        objectstore_props["X-Container-Read"] = '".r:*"'
        self.properties = objectstore_props



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

''' Provide a global configuration for the TOSCA translator'''
import configparser
import os

from toscaparser.utils.gettextutils import _
import translator.common.exception as exception


class ConfigProvider(object):
    '''Global config proxy that wraps a ConfigParser object.

    Allows for class based access to config values. Should only be initialized
    once using the corresponding translator.conf file in the conf directory.

    '''

    # List that captures all of the conf file sections.
    # Append any new sections to this list.
    _sections = ['DEFAULT']
    _translator_config = None

    @classmethod
    def _load_config(cls, conf_file):
        '''Private method only to be called once from the __init__ module'''

        cls._translator_config = configparser.ConfigParser()
        try:
            cls._translator_config.read(conf_file)
        except configparser.ParsingError:
            msg = _('Unable to parse translator.conf file.'
                    'Check to see that it exists in the conf directory.')
            raise exception.ConfFileParseError(message=msg)

    @classmethod
    def get_value(cls, section, key):
        try:
            value = cls._translator_config.get(section, key)
        except configparser.NoOptionError:
            raise exception.ConfOptionNotDefined(key=key, section=section)
        except configparser.NoSectionError:
            raise exception.ConfSectionNotDefined(section=section)

        return value

    @classmethod
    def get_all_values(cls):
        values = []
        for section in cls._sections:
            try:
                values.extend(cls._translator_config.items(section=section))
            except configparser.NoOptionError:
                raise exception.ConfSectionNotDefined(section=section)

        return values

    @classmethod
    def get_translator_logging_file(cls):
        conf_file = ''
        CONF_FILENAME = 'heat_translator_logging.conf'
        conf_path = os.path.dirname(os.path.abspath(__file__))
        conf_file = os.path.join(conf_path, CONF_FILENAME)
        return conf_file



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

''' Initialize the global configuration for the translator '''

import os

from translator.conf.config import ConfigProvider

CONF_FILENAME = 'translator.conf'


def init_global_conf():
    '''Initialize the configuration provider.

    Allows the configuration to be shared throughout the translator code.
    The file used is translator.conf, and is within the conf/ directory. It
    is a standard ini format, and is prcessed using the ConfigParser module.

    '''
    conf_path = os.path.dirname(os.path.abspath(__file__))
    conf_file = os.path.join(conf_path, CONF_FILENAME)
    ConfigProvider._load_config(conf_file)


init_global_conf()



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from translator.hot.syntax.hot_resource import HotResource

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaNetworkPort'
TOSCA_LINKS_TO = 'tosca.relationships.network.LinksTo'
TOSCA_BINDS_TO = 'tosca.relationships.network.BindsTo'


class ToscaNetworkPort(HotResource):
    '''Translate TOSCA node type tosca.nodes.network.Port.'''

    toscatype = 'tosca.nodes.network.Port'

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaNetworkPort, self).__init__(nodetemplate,
                                               type='OS::Neutron::Port',
                                               csar_dir=csar_dir)
        # Default order
        self.order = 0
        pass

    def _generate_networks_for_compute(self, port_resources):
        '''Generate compute networks property list from the port resources.'''
        networks = []
        for resource in port_resources:
            networks.append({'port': '{ get_resource: %s }' % (resource.name)})
        return networks

    def _insert_sorted_resource(self, resources, resource):
        '''Insert a resource in the list of resources and keep the order.'''
        lo = 0
        hi = len(resources)
        while lo < hi:
            mid = (lo + hi) // 2
            if resource.order < resources[mid].order:
                hi = mid
            else:
                lo = mid + 1
        resources.insert(lo, resource)

    def handle_properties(self):
        tosca_props = self.get_tosca_props()
        port_props = {}
        for key, value in tosca_props.items():
            if key == 'ip_address':
                fixed_ip = {}
                fixed_ip['ip_address'] = value
                port_props['fixed_ips'] = [fixed_ip]
            elif key == 'order':
                self.order = value
            # TODO(sdmonov): Need to implement the properties below
            elif key == 'is_default':
                pass
            elif key == 'ip_range_start':
                pass
            elif key == 'ip_range_end':
                pass
            else:
                port_props[key] = value

        links_to = None
        binds_to = None
        for rel, node in self.nodetemplate.relationships.items():
            # Check for LinksTo relations. If found add a network property with
            # the network name into the port
            if not links_to and rel.is_derived_from(TOSCA_LINKS_TO):
                links_to = node

                network_resource = None
                for hot_resource in self.depends_on_nodes:
                    if links_to.name == hot_resource.name:
                        network_resource = hot_resource
                        self.depends_on.remove(hot_resource)
                        break

                if network_resource.existing_resource_id:
                    port_props['network'] =\
                        str(network_resource.existing_resource_id)
                else:
                    port_props['network'] = '{ get_resource: %s }'\
                        % (links_to.name)

            # Check for BindsTo relationship. If found add network to the
            # network property of the corresponding compute resource
            elif not binds_to and rel.is_derived_from(TOSCA_BINDS_TO):
                binds_to = node
                compute_resource = None
                for hot_resource in self.depends_on_nodes:
                    if binds_to.name == hot_resource.name:
                        compute_resource = hot_resource
                        self.depends_on.remove(hot_resource)
                        break
                if compute_resource:
                    port_rsrcs = compute_resource.assoc_port_resources
                    self._insert_sorted_resource(port_rsrcs, self)
                    # TODO(sdmonov): Using generate networks every time we add
                    # a network is not the fastest way to do the things. We
                    # should do this only once at the end.
                    networks = self._generate_networks_for_compute(port_rsrcs)
                    compute_resource.properties['networks'] = networks

        self.properties = port_props



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from toscaparser.tosca_template import ToscaTemplate
from translator.tests.base import TestCase


class ToscaMongoNodejsTest(TestCase):
    parsed_params = {'storage_snapshot_id': 'test_id',
                     'storage_location': '/test', 'cpus': '1',
                     'storage_size': '1'}

    '''TOSCA template with nodejs, app and mongodb on 2 servers.'''
    tosca_tpl = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "../tests/data/tosca_nodejs_mongodb_two_instances.yaml")
    tosca = ToscaTemplate(tosca_tpl, parsed_params)

    def test_relationship_def(self):
        expected_relationship = 'tosca.relationships.HostedOn'
        expected_capabilities_names = 'node'
        for tpl in self.tosca.nodetemplates:
            if tpl.name == 'nodejs':
                def_keys = tpl.type_definition.relationship.keys()
                self.assertIn(
                    expected_relationship,
                    sorted([x.type for x in def_keys]))
                self.assertIn(
                    expected_capabilities_names,
                    sorted([x.capability_name for x in def_keys]))

    def test_relationships(self):
        expected_relationship = ['tosca.relationships.HostedOn']
        expected_relatednodes = ['app_server']
        for tpl in self.tosca.nodetemplates:
            rels = tpl.relationships
            if rels:
                if tpl.name == 'nodejs':
                    self.assertEqual(
                        expected_relationship,
                        sorted([x.type for x in tpl.relationships.keys()]))
                    self.assertEqual(
                        expected_relatednodes,
                        sorted([y.name for y in tpl.relationships.values()]))

    def test_related_nodes(self):
        expected_nodejs = ['app_server']
        actual_nodejs = []
        for tpl in self.tosca.nodetemplates:
            if tpl.name == 'nodejs':
                for node in tpl.related_nodes:
                    actual_nodejs.append(node.name)
        self.assertEqual(sorted(actual_nodejs), expected_nodejs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import json
import os
import shutil
import tempfile
from unittest import mock

from toscaparser.common import exception
from toscaparser.utils.gettextutils import _
import translator.shell as shell
from translator.tests.base import TestCase


class ShellTest(TestCase):
    tosca_helloworld = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data/tosca_helloworld.yaml")
    template_file = '--template-file=' + tosca_helloworld
    template_type = '--template-type=tosca'
    template_validation = "--validate-only"
    failure_msg = _('The program raised an exception unexpectedly.')

    def test_invalid_file_value(self):
        error = self.assertRaises(ValueError,
                                  shell.main, ('--template-file=template.txt',
                                               self.template_type))
        err_msg = _('The path template.txt is not a valid file or URL.')
        self.assertEqual(err_msg, str(error))

    def test_invalid_type_value(self):
        self.assertRaises(SystemExit, shell.main,
                          (self.template_file, '--template-type=xyz'))

    def test_invalid_parameters(self):
        self.assertRaises(ValueError, shell.main,
                          (self.template_file, self.template_type,
                           '--parameters=key'))

    def test_valid_template(self):
        shell.main([self.template_file, self.template_type])

    def test_valid_template_without_type(self):
        try:
            shell.main([self.template_file])
        except Exception:
            self.fail(self.failure_msg)

    def test_valid_template_with_parameters(self):
        tosca_single_instance_wordpress = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "data/tosca_single_instance_wordpress.yaml")
        parameters = '--parameters="cpus=2;db_name=wpdb;db_user=test;'\
                     'db_port=2000;db_root_pwd=fun2test;db_pwd=fun2test"'
        template = '--template-file=' + tosca_single_instance_wordpress
        try:
            shell.main([template, self.template_type, parameters])
        except Exception:
            self.fail(self.failure_msg)

    def test_validate_only(self):
        try:
            shell.main([self.template_file, self.template_type,
                        self.template_validation])
        except Exception:
            self.fail(self.failure_msg)

        template = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "data/tosca_helloworld_invalid.yaml")
        invalid_template = '--template-file=' + template
        self.assertRaises(exception.ValidationError, shell.main,
                          [invalid_template, self.template_type,
                           self.template_validation])

    def test_output_file(self):
        temp_dir = tempfile.mkdtemp()
        temp_file = "/test_translation_output.txt"
        output_file = "--output-file=" + temp_dir + temp_file
        try:
            shell.main([self.template_file, self.template_type, output_file])
        except Exception:
            self.fail(self.failure_msg)
        finally:
            if temp_dir:
                shutil.rmtree(temp_dir)
                self.assertTrue(temp_dir is None or
                                not os.path.exists(temp_dir))

    @mock.patch('uuid.uuid4')
    @mock.patch.object(shell.TranslatorShell, '_create_stack')
    @mock.patch('keystoneauth1.loading.load_auth_from_argparse_arguments')
    @mock.patch('keystoneauth1.loading.load_session_from_argparse_arguments')
    @mock.patch('translator.common.flavors.get_flavors')
    @mock.patch('translator.common.images.get_images')
    def test_template_deploy(self, mock_populate_image_dict,
                             mock_flavor_dict,
                             mock_keystone_session,
                             mock_keystone_auth,
                             mock_client,
                             mock_uuid):
        mock_uuid.return_value = 'abcXXX-abcXXX'
        mock_flavor_dict.return_value = {
            'm1.medium': {'mem_size': 4096, 'disk_size': 40, 'num_cpus': 2}
        }
        mock_populate_image_dict.return_value = {
            "rhel-6.5-test-image": {
                "version": "6.5",
                "architecture": "x86_64",
                "distribution": "RHEL",
                "type": "Linux"
            }
        }

        try:
            data = {
                'outputs': {},
                'heat_template_version': '2013-05-23',
                'description': 'Template for deploying a single server '
                               'with predefined properties.\n',
                'parameters': {},
                'resources': {
                    'my_server': {
                        'type': 'OS::Nova::Server',
                        'properties': {
                            'flavor': 'm1.medium',
                            'user_data_format': 'SOFTWARE_CONFIG',
                            'image': 'rhel-6.5-test-image'
                        }
                    }
                }
            }

            mock_heat_res = {
                "stacks": [
                    {
                        "id": "d648ad27-fb9c-44d1-b293-646ea6c4f8da",
                        "stack_status": "CREATE_IN_PROGRESS",
                    }
                ]
            }

            class mock_response(object):
                def __init__(self, status_code, _content):
                    self.status_code = status_code
                    self._content = _content

            mock_response_obj = mock_response(201, json.dumps(mock_heat_res))
            mock_client.return_value = mock_response_obj
            shell.main([self.template_file, self.template_type, "--deploy"])
            args, kwargs = mock_client.call_args
            self.assertEqual(kwargs["stack_name"],
                             'heat_tosca_helloworld_abcXXX')
            self.assertEqual(kwargs["template"], data)
        except Exception as e:
            self.fail(e)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from toscaparser.common.exception import InvalidPropertyValueError
from translator.hot.syntax.hot_resource import HotResource

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaNetwork'


class ToscaNetwork(HotResource):
    '''Translate TOSCA node type tosca.nodes.network.Network.'''

    toscatype = 'tosca.nodes.network.Network'
    SUBNET_SUFFIX = '_subnet'
    NETWORK_PROPS = ['network_name', 'network_id', 'segmentation_id']
    SUBNET_PROPS = ['ip_version', 'cidr', 'start_ip', 'end_ip', 'gateway_ip']

    existing_resource_id = None

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaNetwork, self).__init__(nodetemplate,
                                           type='OS::Neutron::Net',
                                           csar_dir=csar_dir)
        pass

    def handle_properties(self):
        tosca_props = self.get_tosca_props()

        net_props = {}
        for key, value in tosca_props.items():
            if key in self.NETWORK_PROPS:
                if key == 'network_name':
                    # If CIDR is specified network_name should
                    # be used as the name for the new network.
                    if 'cidr' in tosca_props.keys():
                        net_props['name'] = value
                    # If CIDR is not specified network_name will be used
                    # to lookup existing network. If network_id is specified
                    # together with network_name then network_id should be
                    # used to lookup the network instead
                    elif 'network_id' not in tosca_props.keys():
                        self.hide_resource = True
                        self.existing_resource_id = value
                        break
                elif key == 'network_id':
                    self.hide_resource = True
                    self.existing_resource_id = value
                    break
                elif key == 'segmentation_id':
                    # Hardcode to vxlan for now until we add the network type
                    # and physical network to the spec.
                    net_props['value_specs'] = {'provider:segmentation_id':
                                                value, 'provider:network_type':
                                                'vxlan'}
        self.properties = net_props

    def handle_expansion(self):
        # If the network resource should not be output (they are hidden),
        # there is no need to generate subnet resource
        if self.hide_resource:
            return

        tosca_props = self.get_tosca_props()

        subnet_props = {}

        ip_pool_start = None
        ip_pool_end = None

        for key, value in tosca_props.items():
            if key in self.SUBNET_PROPS:
                if key == 'start_ip':
                    ip_pool_start = value
                elif key == 'end_ip':
                    ip_pool_end = value
                elif key == 'dhcp_enabled':
                    subnet_props['enable_dhcp'] = value
                else:
                    subnet_props[key] = value

        if 'network_id' in tosca_props:
            subnet_props['network'] = tosca_props['network_id']
        else:
            subnet_props['network'] = '{ get_resource: %s }' % (self.name)

        # Handle allocation pools
        # Do this only if both start_ip and end_ip are provided
        # If one of them is missing throw an exception.
        if ip_pool_start and ip_pool_end:
            allocation_pool = {}
            allocation_pool['start'] = ip_pool_start
            allocation_pool['end'] = ip_pool_end
            allocation_pools = [allocation_pool]
            subnet_props['allocation_pools'] = allocation_pools
        elif ip_pool_start:
            raise InvalidPropertyValueError(what=_('start_ip'))
        elif ip_pool_end:
            raise InvalidPropertyValueError(what=_('end_ip'))

        subnet_resource_name = self.name + self.SUBNET_SUFFIX

        hot_resources = [HotResource(self.nodetemplate,
                                     type='OS::Neutron::Subnet',
                                     name=subnet_resource_name,
                                     properties=subnet_props)]
        return hot_resources



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from translator.hot.translate_node_templates import _generate_type_map
from translator.tests.base import TestCase


class TranslateNodeTemplatesTest(TestCase):

    def test_generate_type_map(self):
        expected_type_list = [
            'tosca.nodes.BlockStorage',
            'tosca.nodes.Compute',
            'tosca.nodes.DBMS',
            'tosca.nodes.Database',
            'tosca.nodes.ObjectStorage',
            'tosca.nodes.SoftwareComponent',
            'tosca.nodes.WebApplication',
            'tosca.nodes.WebServer',
            'tosca.nodes.network.FloatingIP',
            'tosca.nodes.network.Network',
            'tosca.nodes.network.Port',
            'tosca.policies.Monitoring',
            'tosca.policies.Placement',
            'tosca.policies.Reservation',
            'tosca.policies.Scaling',
            'tosca.policies.Scaling.Cluster',
            'tosca.nodes.nfv.VNF',
            'tosca.nodes.nfv.Vdu.Compute',
            'tosca.nodes.nfv.Vdu.VirtualBlockStorage',
            'tosca.nodes.nfv.VduCp',
            'tosca.nodes.nfv.VnfVirtualLink',
            'tosca.policies.nfv.InstantiationLevels',
            'tosca.policies.nfv.ScalingAspects',
            'tosca.policies.nfv.VduInitialDelta',
            'tosca.policies.nfv.VduInstantiationLevels',
            'tosca.policies.nfv.VduScalingAspectDeltas',
            'tosca.policies.nfv.VirtualLinkInstantiationLevels'
        ]
        actual_type_list = list(_generate_type_map())

        self.assertCountEqual(expected_type_list, actual_type_list)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from toscaparser.functions import GetInput
from translator.hot.syntax.hot_resource import HotResource

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaBlockStorageAttachment'


class ToscaBlockStorageAttachment(HotResource):
    '''Translate TOSCA relationship AttachesTo for Compute and BlockStorage.'''

    toscatype = 'tosca.nodes.BlockStorageAttachment'

    def __init__(self, template, nodetemplates, instance_uuid, volume_id,
                 csar_dir=None):
        super(ToscaBlockStorageAttachment,
              self).__init__(template, type='OS::Cinder::VolumeAttachment',
                             csar_dir=csar_dir)
        self.nodetemplates = nodetemplates
        self.instance_uuid = {'get_resource': instance_uuid}
        self.volume_id = {'get_resource': volume_id}

    def handle_properties(self):
        tosca_props = {}
        for prop in self.nodetemplate.get_properties_objects():
            if isinstance(prop.value, GetInput):
                tosca_props[prop.name] = {'get_param': prop.value.input_name}
            else:
                tosca_props[prop.name] = prop.value
        self.properties = tosca_props
        # instance_uuid and volume_id for Cinder volume attachment
        self.properties['instance_uuid'] = self.instance_uuid
        self.properties['volume_id'] = self.volume_id
        if 'location' in self.properties:
            self.properties['mountpoint'] = self.properties.pop('location')

        # TOSCA type can have a device name specified,
        # this is unsupported by Heat
        if 'device' in self.properties:
            self.properties.pop('device')

    def handle_life_cycle(self):
        return None, None, None



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
from toscaparser.common.exception import InvalidPropertyValueError
from toscaparser.elements.scalarunit import ScalarUnit_Size
from toscaparser.functions import GetInput
from toscaparser.utils.gettextutils import _
from translator.hot.syntax.hot_resource import HotResource

log = logging.getLogger('heat-translator')

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaBlockStorage'


class ToscaBlockStorage(HotResource):
    '''Translate TOSCA node type tosca.nodes.BlockStorage.'''

    toscatype = 'tosca.nodes.BlockStorage'

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaBlockStorage, self).__init__(nodetemplate,
                                                type='OS::Cinder::Volume',
                                                csar_dir=csar_dir)
        pass

    def handle_properties(self):
        tosca_props = {}
        for prop in self.nodetemplate.get_properties_objects():
            if isinstance(prop.value, GetInput):
                tosca_props[prop.name] = {'get_param': prop.value.input_name}
            else:
                if prop.name == "size":
                    size_value = (ScalarUnit_Size(prop.value).
                                  get_num_from_scalar_unit('GiB'))
                    if size_value == 0:
                        # OpenStack Heat expects size in GB
                        msg = _('Cinder Volume Size unit should be in GB.')
                        log.error(msg)
                        raise InvalidPropertyValueError(
                            what=msg)
                    elif int(size_value) < size_value:
                        size_value = int(size_value) + 1
                        log.warning(_("Cinder unit value should be in "
                                      "multiples of GBs. so corrected "
                                      " %(prop_val)s to %(size_value)s GB.")
                                    % {'prop_val': prop.value,
                                       'size_value': size_value})
                    tosca_props[prop.name] = int(size_value)
                else:
                    tosca_props[prop.name] = prop.value
        self.properties = tosca_props

    def get_hot_attribute(self, attribute, args):
        attr = {}
        # Convert from a TOSCA attribute for a nodetemplate to a HOT
        # attribute for the matching resource.  Unless there is additional
        # runtime support, this should be a one to one mapping.
        if attribute == 'volume_id':
            attr['get_resource'] = self.name
        return attr



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from toscaparser.utils.gettextutils import _
from translator.hot.syntax.hot_template import HotTemplate
from translator.hot.translate_inputs import TranslateInputs
from translator.hot.translate_node_templates import TranslateNodeTemplates
from translator.hot.translate_outputs import TranslateOutputs

log = logging.getLogger('heat-translator')


class TOSCATranslator(object):
    '''Invokes translation methods.'''

    def __init__(self, tosca, parsed_params, deploy=None, csar_dir=None):
        super(TOSCATranslator, self).__init__()
        self.tosca = tosca
        self.hot_template = HotTemplate()
        self.parsed_params = parsed_params
        self.deploy = deploy
        self.csar_dir = csar_dir
        self.node_translator = None
        log.info(_('Initialized parmaters for translation.'))

    def _translate_to_hot_yaml(self):
        self._resolve_input()
        self.hot_template.description = self.tosca.description
        self.hot_template.parameters = self._translate_inputs()
        self.node_translator = TranslateNodeTemplates(self.tosca,
                                                      self.hot_template,
                                                      csar_dir=self.csar_dir)
        self.hot_template.resources = \
            self.node_translator.translate()
        self.hot_template.outputs = self._translate_outputs()
        if self.node_translator.hot_template_version is None:
            self.node_translator.hot_template_version = HotTemplate.LATEST

    def translate(self):
        """Translate to HOT YAML

        This method produces a translated output for main template.
        The nested template, if any referenced by main, will be created
        as a separate file.
        """
        self._translate_to_hot_yaml()

        # TODO(mvelten) go back to calling hot_template.output_to_yaml instead
        # for stdout once embed_substack_templates is correctly implemented
        # return self.hot_template.output_to_yaml(
        #     self.node_translator.hot_template_version)
        yaml_files = self.hot_template.output_to_yaml_files_dict(
            "output.yaml",
            self.node_translator.hot_template_version)
        for name, content in yaml_files.items():
            if name != "output.yaml":
                with open(name, 'w+') as f:
                    f.write(content)

        return yaml_files["output.yaml"]

    def translate_to_yaml_files_dict(self, base_filename):
        """Translate to HOT YAML

        This method produces a translated output containing main and
        any nested templates referenced by main. This output can be
        programmatically stored into different files by using key as
        template name and value as template content.
        """
        self._translate_to_hot_yaml()
        return self.hot_template.output_to_yaml_files_dict(
            base_filename,
            self.node_translator.hot_template_version)

    def _translate_inputs(self):
        translator = TranslateInputs(self.tosca.inputs, self.parsed_params,
                                     self.deploy)
        return translator.translate()

    def _translate_outputs(self):
        translator = TranslateOutputs(self.tosca.outputs, self.node_translator)
        return translator.translate()

    # check all properties for all node and ensure they are resolved
    # to actual value
    def _resolve_input(self):
        for n in self.tosca.nodetemplates:
            for node_prop in n.get_properties_objects():
                if isinstance(node_prop.value, dict) \
                        and 'get_input' in node_prop.value:
                    try:
                        self.parsed_params[node_prop.value['get_input']]
                    except Exception:
                        msg = (_('Must specify all input values in \
                                TOSCA template, missing %s.') %
                               node_prop.value['get_input'])
                        log.error(msg)
                        raise ValueError(msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import OrderedDict
import logging
import os
import yaml

from toscaparser.elements.interfaces import InterfacesDef
from toscaparser.functions import GetInput
from toscaparser.nodetemplate import NodeTemplate
from toscaparser.utils.gettextutils import _


SECTIONS = (TYPE, PROPERTIES, MEDADATA, DEPENDS_ON, UPDATE_POLICY,
            DELETION_POLICY) = \
           ('type', 'properties', 'metadata',
            'depends_on', 'update_policy', 'deletion_policy')

HEAT_TEMPLATE_VERSION = '2013-05-23'
HEAT_DESCRIPTION = 'Scaling template'

policy_type = ['tosca.policies.Placement',
               'tosca.policies.Scaling',
               'tosca.policies.Scaling.Cluster',
               'tosca.policies.Monitoring',
               'tosca.policies.Reservation',
               'tosca.policies.nfv.InstantiationLevels',
               'tosca.policies.nfv.ScalingAspects',
               'tosca.policies.nfv.VduInitialDelta',
               'tosca.policies.nfv.VduInstantiationLevels',
               'tosca.policies.nfv.VduScalingAspectDeltas',
               'tosca.policies.nfv.VirtualLinkInstantiationLevels']
log = logging.getLogger('heat-translator')


class HotResource(object):
    '''Base class for TOSCA node type translation to Heat resource type.'''

    def __init__(self, nodetemplate, name=None, type=None, properties=None,
                 metadata=None, depends_on=None,
                 update_policy=None, deletion_policy=None, csar_dir=None):
        log.debug(_('Translating TOSCA node type to HOT resource type.'))
        self.nodetemplate = nodetemplate
        if name:
            self.name = name
        else:
            self.name = nodetemplate.name
        self.type = type
        self.properties = properties or {}

        self.csar_dir = csar_dir
        # special case for HOT softwareconfig
        cwd = os.getcwd()
        if type == 'OS::Heat::SoftwareConfig':
            config = self.properties.get('config')
            if isinstance(config, dict):
                if self.csar_dir:
                    os.chdir(self.csar_dir)
                    implementation_artifact = os.path.abspath(config.get(
                        'get_file'))
                else:
                    implementation_artifact = config.get('get_file')
                if implementation_artifact:
                    filename, file_extension = os.path.splitext(
                        implementation_artifact)
                    file_extension = file_extension.lower()
                    # artifact_types should be read to find the exact script
                    # type, unfortunately artifact_types doesn't seem to be
                    # supported by the parser
                    if file_extension == '.ansible' \
                            or file_extension == '.yaml' \
                            or file_extension == '.yml':
                        self.properties['group'] = 'ansible'
                    if file_extension == '.pp':
                        self.properties['group'] = 'puppet'

            if self.properties.get('group') is None:
                self.properties['group'] = 'script'
        os.chdir(cwd)
        self.metadata = metadata

        # The difference between depends_on and depends_on_nodes is
        # that depends_on defines dependency in the context of the
        # HOT template and it is used during the template output.
        # Depends_on_nodes defines the direct dependency between the
        # tosca nodes and is not used during the output of the
        # HOT template but for internal processing only. When a tosca
        # node depends on another node it will be always added to
        # depends_on_nodes but not always to depends_on. For example
        # if the source of dependency is a server, the dependency will
        # be added as properties.get_resource and not depends_on
        if depends_on:
            self.depends_on = depends_on
            self.depends_on_nodes = depends_on
        else:
            self.depends_on = []
            self.depends_on_nodes = []
        self.update_policy = update_policy
        self.deletion_policy = deletion_policy
        self.group_dependencies = {}
        # if hide_resource is set to true, then this resource will not be
        # generated in the output yaml.
        self.hide_resource = False

    def handle_properties(self):
        # the property can hold a value or the intrinsic function get_input
        # for value, copy it
        # for get_input, convert to get_param
        for prop in self.nodetemplate.get_properties_objects():
            pass

    def handle_life_cycle(self):
        hot_resources = []
        deploy_lookup = {}
        # TODO(anyone):  sequence for life cycle needs to cover different
        # scenarios and cannot be fixed or hard coded here
        operations_deploy_sequence = ['create', 'configure', 'start']

        operations = HotResource.get_all_operations(self.nodetemplate)

        # create HotResource for each operation used for deployment:
        # create, start, configure
        # ignore the other operations
        # observe the order:  create, start, configure
        # use the current HotResource for the first operation in this order

        # hold the original name since it will be changed during
        # the transformation
        node_name = self.name
        reserve_current = 'NONE'

        for operation in operations_deploy_sequence:
            if operation in operations.keys():
                reserve_current = operation
                break

        # create the set of SoftwareDeployment and SoftwareConfig for
        # the interface operations
        hosting_server = None
        if self.nodetemplate.requirements is not None:
            hosting_server = self._get_hosting_server()

        sw_deployment_resource = HOTSoftwareDeploymentResources(hosting_server)
        server_key = sw_deployment_resource.server_key
        servers = sw_deployment_resource.servers
        sw_deploy_res = sw_deployment_resource.software_deployment

        # hosting_server is None if requirements is None
        hosting_on_server = hosting_server if hosting_server else None
        base_type = HotResource.get_base_type_str(
            self.nodetemplate.type_definition)
        # if we are on a compute node the host is self
        if hosting_on_server is None and base_type == 'tosca.nodes.Compute':
            hosting_on_server = self.name
            servers = {'get_resource': self.name}

        cwd = os.getcwd()
        for operation in operations.values():
            if operation.name in operations_deploy_sequence:
                config_name = node_name + '_' + operation.name + '_config'
                deploy_name = node_name + '_' + operation.name + '_deploy'
                if self.csar_dir:
                    os.chdir(self.csar_dir)
                    get_file = os.path.abspath(operation.implementation)
                else:
                    get_file = operation.implementation
                hot_resources.append(
                    HotResource(self.nodetemplate,
                                config_name,
                                'OS::Heat::SoftwareConfig',
                                {'config':
                                    {'get_file': get_file}},
                                csar_dir=self.csar_dir))
                if operation.name == reserve_current and \
                    base_type != 'tosca.nodes.Compute':
                    deploy_resource = self
                    self.name = deploy_name
                    self.type = sw_deploy_res
                    self.properties = {'config': {'get_resource': config_name},
                                       server_key: servers}
                    deploy_lookup[operation] = self
                else:
                    sd_config = {'config': {'get_resource': config_name},
                                 server_key: servers}
                    deploy_resource = \
                        HotResource(self.nodetemplate,
                                    deploy_name,
                                    sw_deploy_res,
                                    sd_config, csar_dir=self.csar_dir)
                    hot_resources.append(deploy_resource)
                    deploy_lookup[operation] = deploy_resource
                lifecycle_inputs = self._get_lifecycle_inputs(operation)
                if lifecycle_inputs:
                    deploy_resource.properties['input_values'] = \
                        lifecycle_inputs
        os.chdir(cwd)

        # Add dependencies for the set of HOT resources in the sequence defined
        # in operations_deploy_sequence
        # TODO(anyone): find some better way to encode this implicit sequence
        group = {}
        op_index_min = None
        op_index_max = -1
        for op, hot in deploy_lookup.items():
            # position to determine potential preceding nodes
            op_index = operations_deploy_sequence.index(op.name)
            if op_index_min is None or op_index < op_index_min:
                op_index_min = op_index
            if op_index > op_index_max:
                op_index_max = op_index
            for preceding_op_name in \
                    reversed(operations_deploy_sequence[:op_index]):
                preceding_hot = deploy_lookup.get(
                    operations.get(preceding_op_name))
                if preceding_hot:
                    hot.depends_on.append(preceding_hot)
                    hot.depends_on_nodes.append(preceding_hot)
                    group[preceding_hot] = hot
                    break

        if op_index_max >= 0:
            last_deploy = deploy_lookup.get(operations.get(
                operations_deploy_sequence[op_index_max]))
        else:
            last_deploy = None

        # save this dependency chain in the set of HOT resources
        self.group_dependencies.update(group)
        for hot in hot_resources:
            hot.group_dependencies.update(group)

        roles_deploy_resource = self._handle_ansiblegalaxy_roles(
            hot_resources, node_name, servers)

        # add a dependency to this ansible roles deploy to
        # the first "classic" deploy generated for this node
        if roles_deploy_resource and op_index_min:
            first_deploy = deploy_lookup.get(operations.get(
                operations_deploy_sequence[op_index_min]))
            first_deploy.depends_on.append(roles_deploy_resource)
            first_deploy.depends_on_nodes.append(roles_deploy_resource)

        return hot_resources, deploy_lookup, last_deploy

    def _handle_ansiblegalaxy_roles(self, hot_resources, initial_node_name,
                                    hosting_on_server):
        artifacts = self.get_all_artifacts(self.nodetemplate)
        install_roles_script = ''

        sw_deployment_resource = \
            HOTSoftwareDeploymentResources(hosting_on_server)
        server_key = sw_deployment_resource.server_key
        sw_deploy_res = sw_deployment_resource.software_deployment
        for artifact_name, artifact in artifacts.items():
            artifact_type = artifact.get('type', '').lower()
            if artifact_type == 'tosca.artifacts.ansiblegalaxy.role':
                role = artifact.get('file', None)
                if role:
                    install_roles_script += 'ansible-galaxy install ' + role \
                                            + '\n'

        if install_roles_script:
            # remove trailing \n
            install_roles_script = install_roles_script[:-1]
            # add shebang and | to use literal scalar type (for multiline)
            install_roles_script = '|\n#!/bin/bash\n' + install_roles_script

            config_name = initial_node_name + '_install_roles_config'
            deploy_name = initial_node_name + '_install_roles_deploy'
            hot_resources.append(
                HotResource(self.nodetemplate, config_name,
                            'OS::Heat::SoftwareConfig',
                            {'config': install_roles_script},
                            csar_dir=self.csar_dir))
            sd_config = {'config': {'get_resource': config_name},
                         server_key: hosting_on_server}
            deploy_resource = \
                HotResource(self.nodetemplate, deploy_name,
                            sw_deploy_res,
                            sd_config, csar_dir=self.csar_dir)
            hot_resources.append(deploy_resource)

            return deploy_resource

    def handle_connectsto(self, tosca_source, tosca_target, hot_source,
                          hot_target, config_location, operation):
        # The ConnectsTo relationship causes a configuration operation in
        # the target.
        # This hot resource is the software config portion in the HOT template
        # This method adds the matching software deployment with the proper
        # target server and dependency
        if config_location == 'target':
            hosting_server = hot_target._get_hosting_server()
            hot_depends = hot_target
        elif config_location == 'source':
            hosting_server = self._get_hosting_server()
            hot_depends = hot_source
        sw_deployment_resource = HOTSoftwareDeploymentResources(hosting_server)
        server_key = sw_deployment_resource.server_key
        servers = sw_deployment_resource.servers
        sw_deploy_res = sw_deployment_resource.software_deployment

        deploy_name = tosca_source.name + '_' + tosca_target.name + \
            '_connect_deploy'
        sd_config = {'config': {'get_resource': self.name},
                     server_key: servers}
        deploy_resource = \
            HotResource(self.nodetemplate,
                        deploy_name,
                        sw_deploy_res,
                        sd_config,
                        depends_on=[hot_depends], csar_dir=self.csar_dir)
        connect_inputs = self._get_connect_inputs(config_location, operation)
        if connect_inputs:
            deploy_resource.properties['input_values'] = connect_inputs

        return deploy_resource

    def handle_expansion(self):
        pass

    def handle_hosting(self):
        # handle hosting server for the OS:HEAT::SoftwareDeployment
        # from the TOSCA nodetemplate, traverse the relationship chain
        # down to the server
        sw_deploy_group = \
            HOTSoftwareDeploymentResources.HOT_SW_DEPLOYMENT_GROUP_RESOURCE
        sw_deploy = HOTSoftwareDeploymentResources.HOT_SW_DEPLOYMENT_RESOURCE

        if self.properties.get('servers') and \
                self.properties.get('server'):
            del self.properties['server']
        if self.type == sw_deploy_group or self.type == sw_deploy:
            # skip if already have hosting
            # If type is NodeTemplate, look up corresponding HotResrouce
            host_server = self.properties.get('servers') \
                or self.properties.get('server')
            if host_server is None:
                raise Exception(_("Internal Error: expecting host "
                                  "in software deployment"))

            elif isinstance(host_server.get('get_resource'), NodeTemplate):
                self.properties['server']['get_resource'] = \
                    host_server['get_resource'].name

            elif isinstance(host_server, dict) and \
                not host_server.get('get_resource'):
                self.properties['servers'] = \
                    host_server

    def top_of_chain(self):
        dependent = self.group_dependencies.get(self)
        if dependent is None:
            return self
        else:
            return dependent.top_of_chain()

    # this function allows to provides substacks as external files
    # those files will be dumped along the output file.
    #
    # return a dict of filename-content
    def extract_substack_templates(self, base_filename, hot_template_version):
        return {}

    # this function asks the resource to embed substacks
    # into the main template, if any.
    # this is used when the final output is stdout
    def embed_substack_templates(self, hot_template_version):
        pass

    def get_dict_output(self):
        resource_sections = OrderedDict()
        resource_sections[TYPE] = self.type
        if self.properties:
            resource_sections[PROPERTIES] = self.properties
        if self.metadata:
            resource_sections[MEDADATA] = self.metadata
        if self.depends_on:
            resource_sections[DEPENDS_ON] = []
            for depend in self.depends_on:
                resource_sections[DEPENDS_ON].append(depend.name)
        if self.update_policy:
            resource_sections[UPDATE_POLICY] = self.update_policy
        if self.deletion_policy:
            resource_sections[DELETION_POLICY] = self.deletion_policy

        return {self.name: resource_sections}

    def _get_lifecycle_inputs(self, operation):
        # check if this lifecycle operation has input values specified
        # extract and convert to HOT format
        if isinstance(operation.value, str):
            # the operation has a static string
            return {}

        # the operation is a dict {'implemenation': xxx, 'input': yyy}
        inputs = operation.value.get('inputs')
        deploy_inputs = {}
        if inputs:
            for name, value in inputs.items():
                deploy_inputs[name] = value
        return deploy_inputs

    def _get_connect_inputs(self, config_location, operation):
        if config_location == 'target':
            inputs = operation.get('pre_configure_target').get('inputs')
        elif config_location == 'source':
            inputs = operation.get('pre_configure_source').get('inputs')
        deploy_inputs = {}
        if inputs:
            for name, value in inputs.items():
                deploy_inputs[name] = value
        return deploy_inputs

    def _get_hosting_server(self, node_template=None):
        # find the server that hosts this software by checking the
        # requirements and following the hosting chain
        hosting_servers = []
        host_exists = False
        this_node_template = self.nodetemplate \
            if node_template is None else node_template
        for requirement in this_node_template.requirements:
            for requirement_name, assignment in requirement.items():
                for check_node in this_node_template.related_nodes:
                    # check if the capability is Container
                    if isinstance(assignment, dict):
                        node_name = assignment.get('node')
                    else:
                        node_name = assignment
                    if node_name and node_name == check_node.name:
                        if self._is_container_type(requirement_name,
                                                   check_node):
                            hosting_servers.append(check_node.name)
                            host_exists = True
                        elif check_node.related_nodes and not host_exists:
                            return self._get_hosting_server(check_node)
        if hosting_servers:
            return hosting_servers
        return None

    def _is_container_type(self, requirement_name, node):
        # capability is a list of dict
        # For now just check if it's type tosca.nodes.Compute
        # TODO(anyone): match up requirement and capability
        base_type = HotResource.get_base_type_str(node.type_definition)
        if base_type == 'tosca.nodes.Compute':
            return True
        else:
            return False

    def get_hot_attribute(self, attribute, args):
        # this is a place holder and should be implemented by the subclass
        # if translation is needed for the particular attribute
        raise Exception(_("No translation in TOSCA type {0} for attribute "
                          "{1}").format(self.nodetemplate.type, attribute))

    def get_tosca_props(self):
        tosca_props = {}
        for prop in self.nodetemplate.get_properties_objects():
            if isinstance(prop.value, GetInput):
                tosca_props[prop.name] = {'get_param': prop.value.input_name}
            else:
                tosca_props[prop.name] = prop.value
        return tosca_props

    def represent_ordereddict(self, dumper, data):
        nodes = []
        for key, value in data.items():
            node_key = dumper.represent_data(key)
            node_value = dumper.represent_data(value)
            nodes.append((node_key, node_value))
        return yaml.nodes.MappingNode('tag:yaml.org,2002:map', nodes)

    def _handle_nested_template(self, scale_res, yaml_name,
                                hot_template_parameters,
                                parameters=None):
        template_dict = OrderedDict()
        template_dict['heat_template_version'] = HEAT_TEMPLATE_VERSION
        template_dict['description'] = HEAT_DESCRIPTION

        if parameters is not None:
            template_dict['parameters'] = parameters

        if hot_template_parameters:
            all_params = OrderedDict()
            for parameter in hot_template_parameters:
                all_params.update(parameter.get_dict_output())
            template_dict.update({'parameters': all_params})

        template_dict["resources"] = {}
        dict_res = OrderedDict()
        for res in scale_res:
            dict_res = res.get_dict_output()
            res_name = list(dict_res.keys())[0]
            template_dict["resources"][res_name] = \
                dict_res[res_name]

        yaml.add_representer(OrderedDict, self.represent_ordereddict)
        yaml.add_representer(dict, self.represent_ordereddict)
        yaml_string = yaml.dump(template_dict, default_flow_style=False)
        yaml_string = yaml_string.replace('\'', '').replace('\n\n', '\n')
        nested_template = {
            yaml_name: yaml_string
        }
        return nested_template

    def remove_depends_on(self, depends_on_set):
        # Remove all depends_on including depends_on_set.
        for rel, node in self.nodetemplate.relationships.items():
            for do in depends_on_set:
                if rel.is_derived_from(do):
                    for hot_resource in self.depends_on_nodes:
                        if node.name == hot_resource.name and \
                                hot_resource in self.depends_on:
                            self.depends_on.remove(hot_resource)
                            break

    @staticmethod
    def get_all_artifacts(nodetemplate):
        # workaround bug in the parser
        base_type = HotResource.get_base_type_str(nodetemplate.type_definition)
        if base_type in policy_type:
            artifacts = {}
        else:
            artifacts = nodetemplate.type_definition.get_value('artifacts',
                                                               parent=True)
        if not artifacts:
            artifacts = {}
        tpl_artifacts = nodetemplate.entity_tpl.get('artifacts')
        if tpl_artifacts:
            artifacts.update(tpl_artifacts)

        return artifacts

    @staticmethod
    def get_all_operations(node):
        operations = {}
        for operation in node.interfaces:
            operations[operation.name] = operation

        # workaround bug in the parser
        base_type = HotResource.get_base_type_str(node.type_definition)
        if base_type in policy_type:
            return operations

        node_type = node.type_definition

        while True:
            type_operations = HotResource._get_interface_operations_from_type(
                node_type, node, 'Standard')
            type_operations.update(operations)
            operations = type_operations

            if node_type.parent_type is not None:
                node_type = node_type.parent_type
            else:
                return operations

    @staticmethod
    def _get_interface_operations_from_type(node_type, node, lifecycle_name):
        operations = {}
        base_type = HotResource.get_base_type_str(node_type)
        if base_type in policy_type:
            return operations
        if node_type.interfaces and lifecycle_name in node_type.interfaces:
            for name, elems in node_type.interfaces[lifecycle_name].items():
                # ignore empty operations (only type)
                # ignore global interface inputs,
                # concrete inputs are on the operations themselves
                if name != 'type' and name != 'inputs':
                    operations[name] = InterfacesDef(node_type,
                                                     lifecycle_name,
                                                     node, name, elems)
        return operations

    @staticmethod
    def get_base_type_str(node_type):
        if isinstance(node_type, str):
            return node_type
        if node_type.parent_type is not None:
            parent_type_str = None
            if isinstance(node_type.parent_type, str):
                parent_type_str = node_type.parent_type
            else:
                parent_type_str = node_type.parent_type.type

            if parent_type_str and parent_type_str.endswith('.Root'):
                return node_type.type
            return HotResource.get_base_type_str(node_type.parent_type)

        return node_type.type


class HOTSoftwareDeploymentResources(object):
    """Provides HOT Software Deployment resources

    SoftwareDeployment or SoftwareDeploymentGroup Resource
    """

    HOT_SW_DEPLOYMENT_RESOURCE = 'OS::Heat::SoftwareDeployment'
    HOT_SW_DEPLOYMENT_GROUP_RESOURCE = 'OS::Heat::SoftwareDeploymentGroup'

    def __init__(self, hosting_server=None):
        self.software_deployment = self.HOT_SW_DEPLOYMENT_RESOURCE
        self.software_deployment_group = self.HOT_SW_DEPLOYMENT_GROUP_RESOURCE
        self.server_key = 'server'
        self.hosting_server = hosting_server
        self.servers = {}
        if hosting_server is not None:
            if len(self.hosting_server) == 1:
                if isinstance(hosting_server, list):
                    self.servers['get_resource'] = self.hosting_server[0]
            else:
                for server in self.hosting_server:
                    self.servers[server] = {'get_resource': server}
                self.software_deployment = self.software_deployment_group
                self.server_key = 'servers'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

try:
    import novaclient.client
    client_available = True
except ImportError:
    client_available = False
    pass

log = logging.getLogger('heat-translator')


PREDEF_FLAVORS = {
    'm1.xlarge': {'mem_size': 16384, 'disk_size': 160, 'num_cpus': 8},
    'm1.large': {'mem_size': 8192, 'disk_size': 80, 'num_cpus': 4},
    'm1.medium': {'mem_size': 4096, 'disk_size': 40, 'num_cpus': 2},
    'm1.small': {'mem_size': 2048, 'disk_size': 20, 'num_cpus': 1},
    'm1.tiny': {'mem_size': 512, 'disk_size': 1, 'num_cpus': 1},
    'm1.micro': {'mem_size': 128, 'disk_size': 0, 'num_cpus': 1},
    'm1.nano': {'mem_size': 64, 'disk_size': 0, 'num_cpus': 1}
}

SESSION = None

FLAVORS = {}


def get_flavors():
    global FLAVORS

    if FLAVORS:
        return FLAVORS

    if SESSION is not None and client_available:
        try:
            client = novaclient.client.Client("2", session=SESSION)
        except Exception as e:
            # Handles any exception coming from openstack
            log.warn(_('Choosing predefined flavors since received '
                       'Openstack Exception: %s') % str(e))
        else:
            for flv in client.flavors.list(detailed=True):
                FLAVORS[str(flv.name)] = {
                    "mem_size": flv.ram,
                    "disk_size": flv.disk,
                    "num_cpus": flv.vcpus
                }

    if not FLAVORS:
        FLAVORS = PREDEF_FLAVORS

    return FLAVORS



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import defaultdict

from translator.hot.syntax.hot_resource import HotResource
# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaClusterAutoscaling'

SCALE_POLICY = 'senlin.policy.scaling-1.0'
SERVER_TYPE = 'os.nova.server-1.0'
SCALE_TYPE = {'SCALE_IN': 'CLUSTER_SCALE_IN',
              'SCALE_OUT': 'CLUSTER_SCALE_OUT'}

ALARM_METER_NAME = {'utilization': 'cpu_util'}
ALARM_COMPARISON_OPERATOR = {'greater_than': 'gt', 'gerater_equal': 'ge',
                             'less_than': 'lt', 'less_equal': 'le',
                             'equal': 'eq', 'not_equal': 'ne'}
ALARM_STATISTIC = {'mean': 'mean', 'median': 'median', 'summary': 'sum',
                   'maximum': 'max', 'minimum': 'min', 'last': 'last',
                   'std': 'std', 'first': 'first', 'count': 'count'}


class ToscaClusterAutoscaling(HotResource):
    '''Translate TOSCA node type tosca.policies.Scaling.Cluster'''

    toscatype = 'tosca.policies.Scaling.Cluster'

    def __init__(self, policy, csar_dir=None):
        hot_type = "OS::Senlin::Policy"
        super(ToscaClusterAutoscaling, self).__init__(policy,
                                                      type=hot_type,
                                                      csar_dir=csar_dir)
        self.policy = policy

    def _generate_scale_properties(self,
                                   target_cluster_nodes,
                                   cluster_scale_type):
        properties = {}
        bindings = []
        policy_res = {}
        adjustment = {}
        properties["type"] = SCALE_POLICY
        for cluster_node in target_cluster_nodes:
            bindings.append({'cluster': cluster_node})
        properties["bindings"] = bindings
        policy_res["event"] = cluster_scale_type
        adjustment["type"] = "CHANGE_IN_CAPACITY"
        adjustment["number"] = self.\
            policy.entity_tpl["properties"]["increment"]
        policy_res["adjustment"] = adjustment
        properties["properties"] = policy_res
        return properties

    def handle_expansion(self):
        hot_resources = []
        hot_type = 'OS::Aodh::GnocchiAggregationByResourcesAlarm'
        trigger_receivers = defaultdict(list)
        for node in self.policy.targets:
            for trigger in self.policy.entity_tpl['triggers']:
                for action in self.policy.\
                    entity_tpl['triggers'][trigger]['action']:
                    scale_name = action
                    action_sample = self.policy.\
                        entity_tpl['triggers'][trigger]['action'][action]
                    scale_type = action_sample['type']
                    scale_implement = action_sample['implementation']
                (entity, method) = scale_implement.split('.')
                receiver_prop = {}
                receiver_prop['cluster'] = {
                    "get_resource": "%s_cluster" % node
                    }
                receiver_prop['action'] = SCALE_TYPE[scale_type]
                receiver_prop['type'] = method
                receiver_name = node + '_' + scale_name + '_receiver'
                trigger_receivers[trigger].append(receiver_name)
                receiver_resources = HotResource(self.nodetemplate,
                                                 type='OS::Senlin::Receiver',
                                                 name=receiver_name,
                                                 properties=receiver_prop)
                hot_resources.append(receiver_resources)

        for trigger in self.policy.entity_tpl['triggers']:
            sample = self.policy.\
                entity_tpl['triggers'][trigger]['condition']
            (meter_name, comparison_operator, threshold) = \
                sample["constraint"].split()
            threshold = threshold.strip("%")
            alarm_prop = {}
            alarm_prop["description"] = self.policy.entity_tpl['description']
            alarm_prop["metric"] = self.policy.\
                entity_tpl['triggers'][trigger]['event_type']['metric']
            alarm_prop["aggregation_method"] = \
                ALARM_STATISTIC[sample['aggregation_method']]
            alarm_prop["granularity"] = sample["granularity"]
            alarm_prop["evaluation_periods"] = sample["evaluations"]
            alarm_prop["threshold"] = threshold
            alarm_prop["resource_type"] = sample.get("resource_type",
                                                     "instance")
            alarm_prop["comparison_operator"] = \
                ALARM_COMPARISON_OPERATOR[comparison_operator]
            alarm_prop["repeat_actions"] = "True"
            alarm_prop["alarm_actions"] = []
            for index in range(len(trigger_receivers[trigger])):
                alarm_prop["alarm_actions"].\
                    append({'get_attr': [trigger_receivers[trigger][index],
                                         'channel',
                                         'alarm_url']})
            ceilometer_resources = HotResource(self.nodetemplate,
                                               type=hot_type,
                                               name=trigger + '_alarm',
                                               properties=alarm_prop)
            hot_resources.append(ceilometer_resources)
        return hot_resources

    def handle_properties(self, resources):
        remove_resources = []
        networks = defaultdict(list)
        for index, resource in enumerate(resources):
            if resource.type == 'OS::Neutron::Port':
                for hot_resource in resource.depends_on_nodes:
                    if hot_resource.type != 'OS::Neutron::Net':
                        networks[hot_resource.name].\
                            append(
                            {'network': '%s' % resource.properties['network']}
                            )
                remove_resources.append(resource)
            elif resource.type == 'OS::Neutron::Net':
                remove_resources.append(resource)
            elif resource.name in self.policy.targets and \
                resource.type != 'OS::Senlin::Policy':
                props = {}
                del resource.properties['user_data_format']
                del resource.properties['networks']
                props['type'] = SERVER_TYPE
                props['properties'] = resource.properties
                profile_resources = \
                    HotResource(resource,
                                type='OS::Senlin::Profile',
                                name=resource.name,
                                properties=props)
                resources.pop(index)
                resources.insert(index, profile_resources)
        for remove_resource in remove_resources:
            resources.remove(remove_resource)

        for index, resource in enumerate(resources):
            if resource.name in self.policy.targets:
                resource.properties['properties']['networks'] = \
                    networks[resource.name]

        for node in self.policy.targets:
            props = {}
            props["profile"] = {'get_resource': '%s' % node}
            temp = self.policy.entity_tpl["properties"]
            props["min_size"] = temp["min_instances"]
            props["max_size"] = temp["max_instances"]
            props["desired_capacity"] = temp["default_instances"]
            self.cluster_name = '%s_cluster' % node
            cluster_resources = \
                HotResource(self.nodetemplate,
                            type='OS::Senlin::Cluster',
                            name=self.cluster_name,
                            properties=props)
            resources.append(cluster_resources)

        trigger_num = len(self.policy.entity_tpl['triggers'])
        for num, trigger in enumerate(self.policy.entity_tpl['triggers']):
            target_cluster_nodes = []
            for action in self.policy.\
                entity_tpl['triggers'][trigger]['action']:
                scale_type = self.policy.\
                    entity_tpl['triggers'][trigger]['action'][action]['type']
            for node in self.policy.targets:
                target_cluster_nodes.\
                    append({"get_resource": "%s_cluster" % node})
            cluster_scale_type = SCALE_TYPE[scale_type]
            scale_in_props = \
                self._generate_scale_properties(target_cluster_nodes,
                                                cluster_scale_type)
            if num == trigger_num - 1:
                self.name = self.name + '_' + trigger
                self.properties = scale_in_props
                break
            policy_resources = \
                HotResource(self.nodetemplate,
                            type='OS::Senlin::Policy',
                            name=self.name + '_' + trigger,
                            properties=scale_in_props)
            resources.append(policy_resources)
        return resources



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import json
import logging
import math
import numbers
import os
import re
import requests
import tempfile
from urllib.parse import urlparse
import yaml
import zipfile

from toscaparser.utils.gettextutils import _
import toscaparser.utils.yamlparser

YAML_ORDER_PARSER = toscaparser.utils.yamlparser.simple_ordered_parse
log = logging.getLogger('heat-translator')

# Required environment variables to create openstackclient object.
ENV_VARIABLES = ['OS_AUTH_URL', 'OS_PASSWORD', 'OS_USERNAME', 'OS_TENANT_NAME']


class MemoryUnit(object):

    UNIT_SIZE_DEFAULT = 'B'
    UNIT_SIZE_DICT = {'B': 1, 'kB': 1000, 'KiB': 1024, 'MB': 1000000,
                      'MiB': 1048576, 'GB': 1000000000,
                      'GiB': 1073741824, 'TB': 1000000000000,
                      'TiB': 1099511627776}

    @staticmethod
    def convert_unit_size_to_num(size, unit=None):
        """Convert given size to a number representing given unit.

        If unit is None, convert to a number representing UNIT_SIZE_DEFAULT
        :param size: unit size e.g. 1 TB
        :param unit: unit to be converted to e.g GB
        :return: converted number e.g. 1000 for 1 TB size and unit GB
        """
        if unit:
            unit = MemoryUnit.validate_unit(unit)
        else:
            unit = MemoryUnit.UNIT_SIZE_DEFAULT
            log.info(_('A memory unit is not provided for size; using the '
                       'default unit %(default)s.') % {'default': 'B'})
        regex = re.compile(r'(\d*)\s*(\w*)')
        result = regex.match(str(size)).groups()
        if result[1]:
            unit_size = MemoryUnit.validate_unit(result[1])
            converted = int(str_to_num(result[0]) *
                            MemoryUnit.UNIT_SIZE_DICT[unit_size] *
                            math.pow(MemoryUnit.UNIT_SIZE_DICT
                                     [unit], -1))
            log.info(_('Given size %(size)s is converted to %(num)s '
                       '%(unit)s.') % {'size': size,
                     'num': converted, 'unit': unit})
        else:
            converted = (str_to_num(result[0]))
        return converted

    @staticmethod
    def validate_unit(unit):
        if unit in MemoryUnit.UNIT_SIZE_DICT.keys():
            return unit
        else:
            for key in MemoryUnit.UNIT_SIZE_DICT.keys():
                if key.upper() == unit.upper():
                    return key

            msg = _('Provided unit "{0}" is not valid. The valid units are '
                    '{1}').format(unit, MemoryUnit.UNIT_SIZE_DICT.keys())
            log.error(msg)
            raise ValueError(msg)


class CompareUtils(object):

    MISMATCH_VALUE1_LABEL = "<Expected>"
    MISMATCH_VALUE2_LABEL = "<Provided>"
    ORDERLESS_LIST_KEYS = ['allowed_values', 'depends_on']

    @staticmethod
    def compare_dicts(dict1, dict2):
        """Return False if not equal, True if both are equal."""

        if dict1 is None and dict2 is None:
            return True
        if dict1 is None or dict2 is None:
            return False

        both_equal = True
        for dict1_item, dict2_item in zip(dict1.items(), dict2.items()):
            if dict1_item != dict2_item:
                msg = (_("%(label1)s: %(item1)s \n is not equal to \n:"
                         "%(label2)s: %(item2)s")
                       % {'label1': CompareUtils.MISMATCH_VALUE2_LABEL,
                          'item1': dict1_item,
                          'label2': CompareUtils.MISMATCH_VALUE1_LABEL,
                          'item2': dict2_item})
                log.warning(msg)
                both_equal = False
                break
        return both_equal

    @staticmethod
    def compare_hot_yamls(generated_yaml, expected_yaml):
        hot_translated_dict = YAML_ORDER_PARSER(generated_yaml)
        hot_expected_dict = YAML_ORDER_PARSER(expected_yaml)
        return CompareUtils.compare_dicts(hot_translated_dict,
                                          hot_expected_dict)

    @staticmethod
    def reorder(dic):
        '''Canonicalize list items in the dictionary for ease of comparison.

        For properties whose value is a list in which the order does not
        matter, some pre-processing is required to bring those lists into a
        canonical format. We use sorting just to make sure such differences
        in ordering would not cause to a mismatch.
        '''

        if type(dic) is not dict:
            return None

        reordered = {}
        for key in dic.keys():
            value = dic[key]
            if type(value) is dict:
                reordered[key] = CompareUtils.reorder(value)
            elif type(value) is list \
                and key in CompareUtils.ORDERLESS_LIST_KEYS:
                reordered[key] = sorted(value)
            else:
                reordered[key] = value
        return reordered

    @staticmethod
    def diff_dicts(dict1, dict2, reorder=True):
        '''Compares two dictionaries and returns their differences.

        Returns a dictionary of mismatches between the two dictionaries.
        An empty dictionary is returned if two dictionaries are equivalent.
        The reorder parameter indicates whether reordering is required
        before comparison or not.
        '''

        if reorder:
            dict1 = CompareUtils.reorder(dict1)
            dict2 = CompareUtils.reorder(dict2)

        if dict1 is None and dict2 is None:
            return {}
        if dict1 is None or dict2 is None:
            return {CompareUtils.MISMATCH_VALUE1_LABEL: dict1,
                    CompareUtils.MISMATCH_VALUE2_LABEL: dict2}

        diff = {}
        keys1 = set(dict1.keys())
        keys2 = set(dict2.keys())
        for key in keys1.union(keys2):
            if key in keys1 and key not in keys2:
                diff[key] = {CompareUtils.MISMATCH_VALUE1_LABEL: dict1[key],
                             CompareUtils.MISMATCH_VALUE2_LABEL: None}
            elif key not in keys1 and key in keys2:
                diff[key] = {CompareUtils.MISMATCH_VALUE1_LABEL: None,
                             CompareUtils.MISMATCH_VALUE2_LABEL: dict2[key]}
            else:
                val1 = dict1[key]
                val2 = dict2[key]
                if val1 != val2:
                    if type(val1) is dict and type(val2) is dict:
                        diff[key] = CompareUtils.diff_dicts(val1, val2, False)
                    else:
                        diff[key] = {CompareUtils.MISMATCH_VALUE1_LABEL: val1,
                                     CompareUtils.MISMATCH_VALUE2_LABEL: val2}
        return diff


class YamlUtils(object):

    @staticmethod
    def get_dict(yaml_file):
        '''Returns the dictionary representation of the given YAML spec.'''
        try:
            return yaml.safe_load(open(yaml_file))
        except IOError:
            return None

    @staticmethod
    def compare_yamls(yaml1_file, yaml2_file):
        '''Returns true if two dictionaries are equivalent, false otherwise.'''
        dict1 = YamlUtils.get_dict(yaml1_file)
        dict2 = YamlUtils.get_dict(yaml2_file)
        return CompareUtils.compare_dicts(dict1, dict2)

    @staticmethod
    def compare_yaml_dict(yaml_file, dic):
        '''Returns true if yaml matches the dictionary, false otherwise.'''
        return CompareUtils.compare_dicts(YamlUtils.get_dict(yaml_file), dic)


class TranslationUtils(object):

    @staticmethod
    def compare_tosca_translation_with_hot(tosca_file, hot_files, params):
        '''Verify tosca translation against the given hot specification.

        inputs:
        tosca_file: relative local path or URL to the tosca input file
        hot_file: relative path to expected hot output
        params: dictionary of parameter name value pairs

        Returns as a dictionary the difference between the HOT translation
        of the given tosca_file and the given hot_file.
        '''

        from toscaparser.tosca_template import ToscaTemplate
        from translator.hot.tosca_translator import TOSCATranslator

        tosca_tpl = os.path.normpath(os.path.join(
            os.path.dirname(os.path.abspath(__file__)), tosca_file))
        a_file = os.path.isfile(tosca_tpl)
        if not a_file:
            tosca_tpl = tosca_file

        expected_hot_templates = []
        for hot_file in hot_files:
            expected_hot_templates.append(os.path.join(
                os.path.dirname(os.path.abspath(__file__)), hot_file))

        tosca = ToscaTemplate(tosca_tpl, params, a_file)
        translate = TOSCATranslator(tosca, params)

        basename = os.path.basename(hot_files[0])
        output_hot_templates = translate.translate_to_yaml_files_dict(basename)
        output_dict = {}
        for output_hot_template_name in output_hot_templates:
            output_dict[output_hot_template_name] = \
                toscaparser.utils.yamlparser.simple_parse(
                    output_hot_templates[output_hot_template_name])

        expected_output_dict = {}
        for expected_hot_template in expected_hot_templates:
            expected_output_dict[os.path.basename(expected_hot_template)] = \
                YamlUtils.get_dict(expected_hot_template)

        return CompareUtils.diff_dicts(expected_output_dict, output_dict)


class UrlUtils(object):

    @staticmethod
    def validate_url(path):
        """Validates whether the given path is a URL or not.

        If the given path includes a scheme (http, https, ftp, ...) and a net
        location (a domain name such as www.github.com) it is validated as a
        URL.
        """
        parsed = urlparse(path)
        return bool(parsed.scheme) and bool(parsed.netloc)


def str_to_num(value):
    """Convert a string representation of a number into a numeric type."""
    if isinstance(value, numbers.Number) \
            or isinstance(value, int) \
            or isinstance(value, float):
        return value
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return None


def check_for_env_variables():
    return set(ENV_VARIABLES) < set(os.environ.keys())


def get_ks_access_dict():
    tenant_name = os.getenv('OS_TENANT_NAME')
    username = os.getenv('OS_USERNAME')
    password = os.getenv('OS_PASSWORD')
    auth_url = os.getenv('OS_AUTH_URL')

    auth_dict = {
        "auth": {
            "tenantName": tenant_name,
            "passwordCredentials": {
                "username": username,
                "password": password
            }
        }
    }
    headers = {'Content-Type': 'application/json'}
    try:
        keystone_response = requests.post(auth_url + '/tokens',
                                          data=json.dumps(auth_dict),
                                          headers=headers)
        if keystone_response.status_code != 200:
            return None
        return json.loads(keystone_response.content)
    except Exception:
        return None


def get_url_for(access_dict, service_type):
    if access_dict is None:
        return None
    service_catalog = access_dict['access']['serviceCatalog']
    service_url = ''
    for service in service_catalog:
        if service['type'] == service_type:
            service_url = service['endpoints'][0]['publicURL']
            break
    return service_url


def get_token_id(access_dict):
    if access_dict is None:
        return None
    return access_dict['access']['token']['id']


def decompress(zip_file, dir=None):
    """Decompress Zip file

    Decompress any zip file. For example, TOSCA CSAR

    inputs:
       zip_file: file in zip format
       dir: directory to decompress zip. If not provided an unique temporary
            directory will be generated and used.
    return:
       dir: absolute path to the decopressed directory
    """
    if not dir:
        dir = tempfile.NamedTemporaryFile().name
    with zipfile.ZipFile(zip_file, "r") as zf:
        zf.extractall(dir)
    return dir


def get_dict_value(dict_item, key, get_files):
    if key in dict_item:
        return get_files.append(dict_item[key])
    for k, v in dict_item.items():
        if isinstance(v, dict):
            get_dict_value(v, key, get_files)




# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

try:
    import glanceclient.client
    client_available = True
except ImportError:
    client_available = False
    pass

log = logging.getLogger('heat-translator')


PREDEF_IMAGES = {
    'ubuntu-software-config-os-init': {'architecture': 'x86_64',
                                       'os_type': 'linux',
                                       'os_distro': 'ubuntu',
                                       'os_version': '14.04'
                                       },

    'ubuntu-12.04-software-config-os-init': {'architecture': 'x86_64',
                                             'os_type': 'linux',
                                             'os_distro': 'ubuntu',
                                             'os_version': '12.04'
                                             },
    'fedora-amd64-heat-config': {'architecture': 'x86_64',
                                 'os_type': 'linux',
                                 'os_distro': 'fedora',
                                 'os_version': '18.0'
                                 },
    'F18-x86_64-cfntools': {'architecture': 'x86_64',
                            'os_type': 'linux',
                            'os_distro': 'fedora',
                            'os_version': '19'
                            },
    'Fedora-x86_64-20-20131211.1-sda': {'architecture': 'x86_64',
                                        'os_type': 'linux',
                                        'os_distro': 'fedora',
                                        'os_version': '20'
                                        },
    'cirros-0.3.1-x86_64-uec': {'architecture': 'x86_64',
                                'os_type': 'linux',
                                'os_distro': 'cirros',
                                'os_version': '0.3.1'
                                },
    'cirros-0.3.2-x86_64-uec': {'architecture': 'x86_64',
                                'os_type': 'linux',
                                'os_distro': 'cirros',
                                'os_version': '0.3.2'
                                },
    'rhel-6.5-test-image': {'architecture': 'x86_64',
                            'os_type': 'linux',
                            'os_distro': 'rhel',
                            'os_version': '6.5'
                            }
}

SESSION = None

IMAGES = {}


def get_images():
    global IMAGES

    if IMAGES:
        return IMAGES

    if SESSION is not None and client_available:
        try:
            client = glanceclient.client.Client("2", session=SESSION)
        except Exception as e:
            # Handles any exception coming from openstack
            log.warn(_('Choosing predefined images since received '
                       'Openstack Exception: %s') % str(e))
        else:
            for image in client.images.list():
                image_id = image.id.encode('ascii', 'ignore')
                metadata = ["architecture", "type", "distribution", "version",
                            "os_distro", "os_type", "os_version"]
                if any(key in image.keys() for key in metadata):
                    IMAGES[image_id] = {}
                    for key in metadata:
                        if key in image.keys():
                            IMAGES[image_id][key] = image[key]

    if not IMAGES:
        IMAGES = PREDEF_IMAGES

    return IMAGES



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from toscaparser.tests.base import TestCase
import translator.common.utils


class CommonUtilsTest(TestCase):

    MemoryUnit = translator.common.utils.MemoryUnit
    cmpUtils = translator.common.utils.CompareUtils
    yamlUtils = translator.common.utils.YamlUtils
    UrlUtils = translator.common.utils.UrlUtils

    def test_convert_unit_size_to_num(self):
        size = '1 TB'
        num_to_convert = 'GB'
        expected_output = 1000
        output = self.MemoryUnit.convert_unit_size_to_num(size, num_to_convert)
        self.assertEqual(output, expected_output)

        size = '40 GB'
        num_to_convert = 'MB'
        expected_output = 40000
        output = self.MemoryUnit.convert_unit_size_to_num(size, num_to_convert)
        self.assertEqual(output, expected_output)

        size = '20 B'
        num_to_convert = None
        expected_output = 20
        output = self.MemoryUnit.convert_unit_size_to_num(size, num_to_convert)
        self.assertEqual(output, expected_output)

    def test_validate_unit(self):
        unit = 'AB'
        exp_msg = ('Provided unit "{0}" is not valid. The valid units are '
                   '{1}').format(unit, self.MemoryUnit.UNIT_SIZE_DICT.keys())
        try:
            self.MemoryUnit.validate_unit(unit)
        except Exception as err:
            self.assertIsInstance(err, ValueError)
            self.assertEqual(exp_msg, err.__str__())

    def test_unit_size_conversion_to_GNU_standard(self):
        unit = 'gB'
        standard_unit = 'GB'
        converted_unit = self.MemoryUnit.validate_unit(unit)
        self.assertEqual(converted_unit, standard_unit)

        unit = 'KB'
        standard_unit = 'kB'
        converted_unit = self.MemoryUnit.validate_unit(unit)
        self.assertEqual(converted_unit, standard_unit)

        unit = 'kb'
        standard_unit = 'kB'
        converted_unit = self.MemoryUnit.validate_unit(unit)
        self.assertEqual(converted_unit, standard_unit)

        unit = 'kB'
        standard_unit = 'kB'
        converted_unit = self.MemoryUnit.validate_unit(unit)
        self.assertEqual(converted_unit, standard_unit)

        unit = 'MIB'
        standard_unit = 'MiB'
        converted_unit = self.MemoryUnit.validate_unit(unit)
        self.assertEqual(converted_unit, standard_unit)

    def test_str_to_num_value_error(self):
        str_to_convert = '55063.000000'
        expected_output = 55063.0
        output = translator.common.utils.str_to_num(str_to_convert)
        self.assertEqual(output, expected_output)

    def test_compare_dicts_unequal(self):
        dict1 = {'allowed_values': [1, 2, 4, 8],
                 'server3': {'depends_on': ['server1', 'server2']}}
        dict2 = {'allowed_values': [1, 2, 4, 8],
                 'server3': {'depends_on': ['server2', 'server1']}}
        self.assertFalse(self.cmpUtils.compare_dicts(dict1, dict2))

    def test_dicts_equivalent_empty_dicts(self):
        self.assertTrue(self.cmpUtils.compare_dicts(None, None))
        self.assertFalse(self.cmpUtils.compare_dicts(None, {}))
        self.assertFalse(self.cmpUtils.compare_dicts(None, {'x': '2'}))

    def test_compareutils_reorder(self):
        dic = {'output': {'website_url': {'value': {'get_attr':
                                                    ['server', 'networks',
                                                     'private', 0]}}},
               'allowed_values': [2, 8, 1, 4],
               'server3': {'depends_on': ['server2', 'server1']}}
        reordered_dic = {'output': {'website_url': {'value': {'get_attr':
                                                    ['server', 'networks',
                                                     'private', 0]}}},
                         'allowed_values': [1, 2, 4, 8],
                         'server3': {'depends_on': ['server1', 'server2']}}
        self.assertEqual(reordered_dic, self.cmpUtils.reorder(dic))

    def test_compareutils_diff_dicts_both_null(self):
        expected = None
        provided = None
        self.assertEqual({},
                         self.cmpUtils.diff_dicts(expected, provided))

    def test_compareutils_diff_dicts_one_null(self):
        expected = {'keyname': 'userkey'}
        provided = None
        self.assertEqual(
            {self.cmpUtils.MISMATCH_VALUE1_LABEL: {'keyname': 'userkey'},
             self.cmpUtils.MISMATCH_VALUE2_LABEL: None},
            self.cmpUtils.diff_dicts(expected, provided))

    def test_compareutils_diff_dicts_missing_key(self):
        expected = {'server3': {'depends_on': ['server1', 'server2'],
                                'keyname': 'userkey'}}
        provided = {'server3': {'depends_on': ['server2', 'server1']}}
        self.assertEqual(
            {'server3': {'keyname':
             {self.cmpUtils.MISMATCH_VALUE1_LABEL: 'userkey',
              self.cmpUtils.MISMATCH_VALUE2_LABEL: None}}},
            self.cmpUtils.diff_dicts(expected, provided))

    def test_compareutils_diff_dicts_missing_key_other_dict(self):
        expected = {'server3': {'depends_on': ['server1', 'server2']}}
        provided = {'server3': {'depends_on': ['server2', 'server1'],
                                'keyname': 'userkey'}}
        self.assertEqual(
            {'server3': {'keyname':
             {self.cmpUtils.MISMATCH_VALUE1_LABEL: None,
              self.cmpUtils.MISMATCH_VALUE2_LABEL: 'userkey'}}},
            self.cmpUtils.diff_dicts(expected, provided))

    def test_compareutils_diff_dicts_value_diff(self):
        expected = \
            {'output':
             {'website_url':
              {'value':
               {'get_attr': ['server', 'networks', 'private', 0]}}},
             'server3': {'depends_on': ['server2', 'server1']}}
        provided = \
            {'output':
             {'website_url':
              {'value':
               {'get_attr': ['server', 'networks', 'public', 0]}}},
             'server3': {'depends_on': ['server2', 'server1']}}
        self.assertEqual(
            {'output':
             {'website_url':
              {'value':
               {'get_attr':
                {self.cmpUtils.MISMATCH_VALUE1_LABEL:
                 ['server', 'networks', 'private', 0],
                 self.cmpUtils.MISMATCH_VALUE2_LABEL:
                 ['server', 'networks', 'public', 0]}}}}},
            self.cmpUtils.diff_dicts(expected, provided))

    def test_yamlutils_get_dict_missing_file(self):
        self.assertIsNone(self.yamlUtils.get_dict('./no_file.yaml'))

    def test_yamlutils_get_dict(self):
        yaml_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../tests/data/custom_types/rsyslog.yaml')
        dict = \
            {'tosca_definitions_version': 'tosca_simple_yaml_1_0',
             'description':
             'RSYSLOG is the Rocket-fast SYStem for LOG processing.\n',
             'node_types':
             {'tosca.nodes.SoftwareComponent.Rsyslog':
              {'derived_from': 'tosca.nodes.SoftwareComponent',
               'requirements':
               [{'log_endpoint':
                 {'capability': 'tosca.capabilities.Endpoint',
                  'node': 'tosca.nodes.SoftwareComponent.Logstash',
                  'relationship': 'tosca.relationships.ConnectsTo'}}]}}}
        self.assertEqual(dict, self.yamlUtils.get_dict(yaml_file))

    def test_yamlutils_compare_yamls(self):
        yaml_file1 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../tests/data/custom_types/kibana.yaml')
        yaml_file2 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../tests/data/custom_types/collectd.yaml')
        self.assertTrue(self.yamlUtils.compare_yamls(yaml_file1, yaml_file1))
        self.assertFalse(self.yamlUtils.compare_yamls(yaml_file1, yaml_file2))

    def test_yamlutils_compare_yaml_dict(self):
        yaml_file1 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../tests/data/custom_types/rsyslog.yaml')
        yaml_file2 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../tests/data/custom_types/collectd.yaml')
        dict = \
            {'tosca_definitions_version': 'tosca_simple_yaml_1_0',
             'description':
             'RSYSLOG is the Rocket-fast SYStem for LOG processing.\n',
             'node_types':
             {'tosca.nodes.SoftwareComponent.Rsyslog':
              {'derived_from': 'tosca.nodes.SoftwareComponent',
               'requirements':
               [{'log_endpoint':
                 {'capability': 'tosca.capabilities.Endpoint',
                  'node': 'tosca.nodes.SoftwareComponent.Logstash',
                  'relationship': 'tosca.relationships.ConnectsTo'}}]}}}
        self.assertEqual({}, self.cmpUtils.diff_dicts(
            self.yamlUtils.get_dict(yaml_file1), dict))
        self.assertFalse(self.yamlUtils.compare_yaml_dict(yaml_file2, dict))

    def test_assert_value_is_num(self):
        value = 1
        output = translator.common.utils.str_to_num(value)
        self.assertEqual(value, output)

    def test_urlutils_validate_url(self):
        self.assertTrue(self.UrlUtils.validate_url("http://www.github.com/"))
        self.assertTrue(
            self.UrlUtils.validate_url("https://github.com:81/a/2/a.b"))
        self.assertTrue(self.UrlUtils.validate_url("ftp://github.com"))
        self.assertFalse(self.UrlUtils.validate_url("github.com"))
        self.assertFalse(self.UrlUtils.validate_url("123"))
        self.assertFalse(self.UrlUtils.validate_url("a/b/c"))

    def test_get_dict_value(self):
        single_snippet = \
            {'nodejs_create_config':
             {'type': 'tosca.nodes.SoftwareConfig',
              'properties':
              {'config':
               {'get_file': 'create.sh'}}}}
        actual_output_single_snippet = []
        ex_output_single_snippet = ['create.sh']
        translator.common.utils.get_dict_value(single_snippet, "get_file",
                                               actual_output_single_snippet)
        self.assertEqual(actual_output_single_snippet,
                         ex_output_single_snippet)
        multi_snippet = \
            {'resources':
             {'nodejs_create_config':
              {'type': 'tosca.nodes.SoftwareConfig',
               'properties':
               {'config':
                {'get_file': 'nodejs/create.sh'}}},
              'mongodb_create_config':
              {'type': 'tosca.nodes.SoftwareConfig',
               'properties':
               {'config':
                {'get_file': 'mongodb/create.sh'}}}}}

        actual_output_multi_snippet = []
        ex_output_multi_snippet = ['mongodb/create.sh',
                                   'nodejs/create.sh']
        translator.common.utils.get_dict_value(multi_snippet, "get_file",
                                               actual_output_multi_snippet)
        self.assertEqual(sorted(actual_output_multi_snippet),
                         ex_output_multi_snippet)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from translator.common.utils import MemoryUnit
from translator.hot.syntax.hot_resource import HotResource


# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaNfvVduVirtualBlockStorage'


class ToscaNfvVduVirtualBlockStorage(HotResource):
    """Translate TOSCA node type tosca.nodes.nfv.Vdu.VirtualBlockStorage."""

    toscatype = 'tosca.nodes.nfv.Vdu.VirtualBlockStorage'
    IMAGE = '#ADD_YOUR_IMAGE_HERE'

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaNfvVduVirtualBlockStorage, self).__init__(
            nodetemplate,
            type='OS::Cinder::Volume',
            csar_dir=csar_dir)

    def handle_properties(self):
        tosca_props = self.get_tosca_props()
        own_props = {}

        for key, value in tosca_props.items():
            # This image's value is replaced after translation.
            if key == 'sw_image_data':
                # Users have to add an image ID or name because heat deprecated
                # creating an image from a local file.
                own_props['image'] = '%s' % (
                    self.IMAGE
                )
            elif key == 'virtual_block_storage_data':
                # Convert to GiB
                own_props['size'] = \
                    MemoryUnit.convert_unit_size_to_num(
                        value['size_of_storage'], 'GiB')

        self.properties = own_props



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.nodetemplate import NodeTemplate
from toscaparser.tests.base import TestCase
from toscaparser.utils.gettextutils import _
import toscaparser.utils.yamlparser
from translator.hot.tosca.tosca_object_storage import ToscaObjectStorage


class ToscaObjectStoreTest(TestCase):

    def _tosca_objectstore_test(self, tpl_snippet, expectedprops):
        nodetemplates = (toscaparser.utils.yamlparser.
                         simple_parse(tpl_snippet)['node_templates'])
        name = list(nodetemplates.keys())[0]
        try:
            nodetemplate = NodeTemplate(name, nodetemplates)
            tosca_object_store = ToscaObjectStorage(nodetemplate)
            tosca_object_store.handle_properties()
            if not self._compare_properties(tosca_object_store.properties,
                                            expectedprops):
                raise Exception(_("Hot Properties are not"
                                  " same as expected properties"))
        except Exception:
            # for time being rethrowing. Will be handled future based
            # on new development
            raise

    def _compare_properties(self, hotprops, expectedprops):
        return all(item in hotprops.items() for item in expectedprops.items())

    def test_node_objectstorage_with_properties(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.ObjectStorage
            properties:
              name: test
              size: 1024 KB
              maxsize: 1 MB
        '''
        expectedprops = {'name': 'test',
                         'X-Container-Meta': {'Quota-Bytes': 1000000}}
        self._tosca_objectstore_test(
            tpl_snippet,
            expectedprops)

    def test_node_objectstorage_with_few_properties(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.ObjectStorage
            properties:
              name: test
              size: 1024 B
        '''
        expectedprops = {'name': 'test',
                         'X-Container-Meta': {'Quota-Bytes': 1024}}
        self._tosca_objectstore_test(
            tpl_snippet,
            expectedprops)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import yaml

from translator.common.utils import MemoryUnit
from translator.hot.syntax.hot_resource import HotResource
log = logging.getLogger('heat-translator')


# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaNfvVduCompute'
TOSCA_ATTACHES_TO = 'tosca.relationships.nfv.AttachesTo'
depends_on_set = (
    TOSCA_ATTACHES_TO,
)


class ToscaNfvVduCompute(HotResource):
    """Translate TOSCA node type tosca.nodes.nfv.Vdu.Compute."""

    toscatype = 'tosca.nodes.nfv.Vdu.Compute'
    IMAGE = '#ADD_YOUR_IMAGE_HERE'
    FLAVOR_SUFFIX = '_flavor'

    def __init__(self, nodetemplate, csar_dir=None):
        super(ToscaNfvVduCompute, self).__init__(
            nodetemplate,
            type='OS::Nova::Server',
            csar_dir=csar_dir)
        # Associated resources
        self.assoc_port_resources = []
        self.virtual_storages = []
        self.flavor_resource_name = None

    def handle_properties(self):
        tosca_props = self.get_tosca_props()
        own_props = {}

        own_props['flavor'] = '{ get_resource: %s%s }' % (
            self.name,
            self.FLAVOR_SUFFIX,
        )

        for key, value in tosca_props.items():
            if key == 'name':
                own_props['name'] = value
            elif key == 'sw_image_data':
                # Users have to add an image ID or name because heat deprecated
                # creating an image from a local file.
                own_props['image'] = '%s' % (
                    self.IMAGE
                )

        # Multi virtual_storages support
        requirements = self.nodetemplate.requirements
        self.virtual_storages = []
        for req in requirements:
            vs_val = req.get('virtual_storage')
            if isinstance(vs_val, dict):
                if vs_val.get('node') is not None:
                    self.virtual_storages.append(vs_val['node'])
            else:
                self.virtual_storages.append(vs_val)
        dict_rsrcs = self.nodetemplate.templates
        bdmv2s = []
        for vs_name in self.virtual_storages:
            # Supported virtual storage types:
            # - tosca.nodes.nfv.Vdu.VirtualBlockStorage
            vs_type = None
            for rsrc_name, rsrc_info in dict_rsrcs.items():
                rsrc_type = rsrc_info['type']
                if rsrc_type == 'tosca.nodes.nfv.Vdu.VirtualBlockStorage':
                    vs_type = rsrc_type

            if vs_type is not None:
                bdmv2s.append({
                    'volume_id': '{ get_resource: %s }' % (vs_name)
                })
            else:
                log.warning(('Unsupported virtual_storage, '
                             'vdu_name:%s, virtual_storage_name:%s')
                            % (self.name, vs_name))

        if bdmv2s != []:
            own_props['block_device_mapping_v2'] = bdmv2s

        self.flavor_resource_name = self.name + self.FLAVOR_SUFFIX

        # Remove depends_on
        self.remove_depends_on(depends_on_set)

        self.properties = own_props

    def handle_expansion(self):
        hot_resources = []
        flavor_props = {}

        tosca_capas = self.nodetemplate.get_capability('virtual_compute')
        tosca_c_props = tosca_capas.get_properties_objects()

        for prop in tosca_c_props:
            if prop.name == 'virtual_cpu':
                flavor_props['vcpus'] = prop.value['num_virtual_cpu']
            elif prop.name == 'virtual_memory':
                # Convert to MiB
                flavor_props['ram'] = \
                    MemoryUnit.convert_unit_size_to_num(
                        prop.value['virtual_mem_size'], 'MiB')
            elif prop.name == 'virtual_local_storage':
                # Convert to GiB
                flavor_props['disk'] = \
                    MemoryUnit.convert_unit_size_to_num(
                        prop.value[0]['size_of_storage'], 'GiB')
            elif prop.name == 'compute_requirements':
                if prop.value.get('nova_extra_specs') is not None:
                    nova_es = prop.value['nova_extra_specs']
                    try:
                        # Check if it is in yaml format
                        yaml.safe_load(nova_es)
                        flavor_props['extra_specs'] = nova_es
                    except Exception:
                        log.warning(('Unsupported format of '
                                     'compute_requirements, '
                                     'vdu_name:%s, nova_extra_specs:%s')
                                    % (self.name, nova_es))

        hot_resources.append(
            HotResource(
                self.nodetemplate,
                type='OS::Nova::Flavor',
                name=self.flavor_resource_name,
                properties=flavor_props,
            )
        )

        return hot_resources



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from translator.hot.syntax.hot_resource import HotResource

# Name used to dynamically load appropriate map class.
TARGET_CLASS_NAME = 'ToscaReservation'

log = logging.getLogger('heat-translator')


class ToscaReservation(HotResource):
    '''Translate TOSCA node type tosca.policies.Reservation'''

    toscatype = 'tosca.policies.Reservation'

    def __init__(self, policy, csar_dir=None):
        hot_type = "OS::Aodh::EventAlarm"
        super(ToscaReservation, self).__init__(policy,
                                               type=hot_type,
                                               csar_dir=csar_dir)
        self.policy = policy
        self.filter = list()

    def handle_expansion(self):
        extra_resources = list()
        hot_type = 'OS::Aodh::EventAlarm'
        for action_name in self.filter:
            prop = self._get_event_type(action_name)
            resrv_resources = HotResource(self.nodetemplate,
                                          type=hot_type,
                                          name=action_name,
                                          properties=prop)
            extra_resources.append(resrv_resources)
        return extra_resources

    def handle_properties(self):
        if self.policy.entity_tpl.get('reservation'):
            reservation_prop = self.policy.entity_tpl["reservation"].items()
            for action_name, action_dict in reservation_prop:
                if action_name != 'properties':
                    self.filter.append(action_name)
            self.name = self.filter[0]

    def _get_event_type(self, action):
        if action == 'start_actions':
            return {'event_type': 'lease.event.start_lease'}
        if action == 'before_end_actions':
            return {'event_type': 'lease.event.before_end_lease'}
        if action == 'end_actions':
            return {'event_type': 'lease.event.end_lease'}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.nodetemplate import NodeTemplate
from toscaparser.policy import Policy
from toscaparser.tests.base import TestCase
import toscaparser.utils.yamlparser
from translator.hot.tosca.tosca_compute import ToscaCompute
from translator.hot.tosca.tosca_policies import ToscaPolicies


class ToscaPoicyTest(TestCase):

    def _tosca_policy_test(self, tpl_snippet, expectedprops):
        nodetemplates = (toscaparser.utils.yamlparser.
                         simple_parse(tpl_snippet)['node_templates'])
        policies = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet)['policies'])
        name = list(nodetemplates.keys())[0]
        policy_name = list(policies[0].keys())[0]
        for policy in policies:
            tpl = policy[policy_name]
            targets = tpl["targets"]
        try:
            nodetemplate = NodeTemplate(name, nodetemplates)
            toscacompute = ToscaCompute(nodetemplate)
            toscacompute.handle_properties()

            # adding a property to test that
            # ToscaPolicies.handle_properties does not overwrite this.
            toscacompute.properties['scheduler_hints'] = {
                'target_cell': 'cell0'}

            policy = Policy(policy_name, tpl, targets,
                            "node_templates")
            toscapolicy = ToscaPolicies(policy)
            nodetemplate = [toscacompute]
            toscapolicy.handle_properties(nodetemplate)

            self.assertEqual(toscacompute.properties, expectedprops)
        except Exception:
            raise

    def test_compute_with_policies(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            capabilities:
              host:
                properties:
                  disk_size: 10 GB
                  num_cpus: 4
                  mem_size: 4 GB
              os:
                properties:
                  architecture: x86_64
                  type: Linux
                  distribution: Fedora
                  version: 18.0
        policies:
          - my_compute_placement_policy:
              type: tosca.policies.Placement
              description: Apply my placement policy to my application servers
              targets: [ server ]
        '''
        expectedprops = {'flavor': 'm1.large',
                         'image': 'fedora-amd64-heat-config',
                         'scheduler_hints': {
                             'group': {
                                 'get_resource':
                                 'my_compute_placement_policy'},
                             'target_cell': 'cell0'},
                         'user_data_format': 'SOFTWARE_CONFIG'}
        self._tosca_policy_test(
            tpl_snippet,
            expectedprops)
